# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

from copy import deepcopy
from datetime import timedelta
from threading import local

from cryptography import x509
from cryptography.hazmat.backends import default_backend

from . import ca_settings
from .extensions import KEY_TO_EXTENSION
from .extensions import AuthorityInformationAccess
from .extensions import AuthorityKeyIdentifier
from .extensions import BasicConstraints
from .extensions import CRLDistributionPoints
from .extensions import DistributionPoint
from .extensions import Extension
from .extensions import IssuerAlternativeName
from .extensions import SubjectAlternativeName
from .extensions import SubjectKeyIdentifier
from .signals import pre_issue_cert
from .subject import Subject
from .utils import get_cert_builder
from .utils import parse_general_name
from .utils import parse_hash_algorithm
from .utils import shlex_split


class Profile(object):
    """A certificate profile defining properties and extensions of a certificate.

    Instances of this class usually represent profiles defined in :ref:`CA_PROFILES <settings-ca-profiles>`,
    but you can also create your own profile to create a different type of certificate. An instance of this
    class can be used to create a signed certificate based on the given CA::

        >>> Profile('example', subject='/C=AT', extensions={'ocsp_no_check': {}})
        <Profile: 'example'>
    """

    def __init__(self, name, subject=None, algorithm=None, extensions=None, cn_in_san=True, expires=None,
                 issuer_name=None, description='', autogenerated=False,
                 add_crl_url=True, add_ocsp_url=True, add_issuer_url=True, add_issuer_alternative_name=True,
                 **kwargs):
        self.name = name

        # self.subject is default subject with updates from subject argument
        if subject is not None:
            if not isinstance(subject, Subject):
                subject = Subject(subject)  # NOTE: also accepts None
            self.subject = subject
        else:
            self.subject = Subject(ca_settings.CA_DEFAULT_SUBJECT)

        self.algorithm = parse_hash_algorithm(algorithm)
        self.extensions = deepcopy(extensions) or {}
        self.cn_in_san = cn_in_san
        self.expires = expires or ca_settings.CA_DEFAULT_EXPIRES
        self.issuer_name = issuer_name
        self.add_crl_url = add_crl_url
        self.add_issuer_url = add_issuer_url
        self.add_ocsp_url = add_ocsp_url
        self.add_issuer_alternative_name = add_issuer_alternative_name
        self.description = description
        self.autogenerated = autogenerated

        if isinstance(self.expires, int):
            self.expires = timedelta(days=self.expires)

        # cast extensions to their respective classes
        for key, extension in self.extensions.items():
            if not isinstance(extension, Extension):
                self.extensions[key] = KEY_TO_EXTENSION[key](extension)

        # set some sane extension defaults
        self.extensions.setdefault(BasicConstraints.key, BasicConstraints())

    def __eq__(self, o):
        if isinstance(o, (Profile, DefaultProfileProxy)) is False:
            return False
        algo = isinstance(o.algorithm, type(self.algorithm))

        return self.name == o.name and self.subject == o.subject and \
            algo and self.extensions == o.extensions and \
            self.cn_in_san == o.cn_in_san and self.expires == o.expires and \
            self.issuer_name == o.issuer_name and self.add_crl_url == o.add_crl_url and \
            self.add_issuer_url == o.add_issuer_url and self.add_ocsp_url == o.add_ocsp_url and \
            self.add_issuer_alternative_name == o.add_issuer_alternative_name and \
            self.description == o.description

    def _parse_extension_value(self, key, value):
        """Parse an extension value into a django_ca extension."""

        if isinstance(value, Extension):
            return value
        elif value is None:
            return None
        else:
            return KEY_TO_EXTENSION[key](value)

    def __repr__(self):
        return '<Profile: %r>' % self.name

    def __str__(self):
        return repr(self)

    def create_cert(self, ca, csr, subject=None, expires=None, algorithm=None, extensions=None,
                    cn_in_san=None, add_crl_url=None, add_ocsp_url=None, add_issuer_url=None,
                    add_issuer_alternative_name=None, password=None):
        """Create a x509 certificate based on this profile, the passed CA and input parameters.

        This function is the core function used to create x509 certificates. In it's simplest form, you only
        need to pass a ca, a csr and a subject to get a valid certificate::

            >>> profile = get_profile('webserver')
            >>> profile.create_cert(ca, csr, subject='/CN=example.com')  # doctest: +ELLIPSIS
            <Certificate(subject=<Name(...,CN=example.com)>, ...)>

        The function will add CRL, OCSP, Issuer and IssuerAlternativeName URLs based on the CA if the profile
        has the *add_crl_url*, *add_ocsp_url* and *add_issuer_url* and *add_issuer_alternative_name* values
        set. Parameters to this function with the same name allow you override this behaviour.

        The function allows you to override profile values using the *expires* and *algorithm* values. You can
        pass additional *extensions* as a list, which will override any extensions from the profile, but the
        CA passed will append to these extensions unless the *add_...* values are ``False``.

        Parameters
        ----------

        ca : :py:class:`~django_ca.models.CertificateAuthority`
            The CA to sign the certificate with.
        csr : str or :py:class:`~cg:cryptography.x509.CertificateSigningRequest`
            The CSR for the certificate.
        subject : dict or str or :py:class:`~django_ca.subject.Subject`
            Update the subject string, e.g. ``"/CN=example.com"`` or ``Subject("/CN=example.com")``. The
            values from the passed subject will update the profiles subject.
        expires : timedelta, optional
            Override when this certificate will expire.
        algorithm : str or :py:class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm`, optional
            Override the hash algorithm used when signing the certificate, passed to
            :py:func:`~django_ca.utils.parse_hash_algorithm`.
        extensions : list or dict of :py:class:`~django_ca.extensions.Extension`
            List or dict of additional extensions to set for the certificate. Note that values from the CA
            might update the passed extensions: For example, if you pass an
            :py:class:`~django_ca.extensions.IssuerAlternativeName` extension, *add_issuer_alternative_name*
            is ``True`` and the passed CA has an IssuerAlternativeName set, that value will be appended to the
            extension you pass here. If you pass a dict with a ``None`` value, that extension will be removed
            from the profile.
        cn_in_san : bool, optional
            Override if the CommonName should be added as an SubjectAlternativeName. If not passed, the value
            set in the profile is used.
        add_crl_url : bool, optional
            Override if any CRL URLs from the CA should be added to the CA. If not passed, the value set in
            the profile is used.
        add_ocsp_url : bool, optional
            Override if any OCSP URLs from the CA should be added to the CA. If not passed, the value set in
            the profile is used.
        add_issuer_url : bool, optional
            Override if any Issuer URLs from the CA should be added to the CA. If not passed, the value set in
            the profile is used.
        add_issuer_alternative_name : bool, optional
            Override if any IssuerAlternativeNames from the CA should be added to the CA. If not passed, the
            value set in the profile is used.
        password: bytes or str, optional
            The password to the private key of the CA.

        Returns
        -------

        cryptography.x509.Certificate
            The signed certificate.
        """

        # Compute default values
        if extensions is None:
            extensions = {}
        elif isinstance(extensions, dict):
            extensions = {k: self._parse_extension_value(k, v) for k, v in extensions.items()}
        else:
            extensions = {e.key: e for e in extensions}

        # Get overrides values from profile if not passed as parameter
        if cn_in_san is None:
            cn_in_san = self.cn_in_san
        if add_crl_url is None:
            add_crl_url = self.add_crl_url
        if add_ocsp_url is None:
            add_ocsp_url = self.add_ocsp_url
        if add_issuer_url is None:
            add_issuer_url = self.add_issuer_url
        if add_issuer_alternative_name is None:
            add_issuer_alternative_name = self.add_issuer_alternative_name

        cert_extensions = deepcopy(self.extensions)
        cert_extensions.update(extensions)
        cert_extensions = {k: v for k, v in cert_extensions.items() if v is not None}
        cert_subject = deepcopy(self.subject)

        issuer_name = self._update_from_ca(
            ca, cert_extensions, add_crl_url=add_crl_url, add_ocsp_url=add_ocsp_url,
            add_issuer_url=add_issuer_url, add_issuer_alternative_name=add_issuer_alternative_name)

        if not isinstance(subject, Subject):
            subject = Subject(subject)  # NOTE: also accepts None
        cert_subject.update(subject)

        if expires is None:
            expires = self.expires
        if algorithm is None:
            algorithm = self.algorithm
        else:
            algorithm = parse_hash_algorithm(algorithm)

        # Finally, update SAN with the current CN, if set and requested
        self._update_san_from_cn(cn_in_san, subject=cert_subject, extensions=cert_extensions)
        # TODO: fail if there is no CN and no SANa

        if not subject.get('CN') and (SubjectAlternativeName.key not in extensions or not
                                      extensions[SubjectAlternativeName.key].value):
            raise ValueError("Must name at least a CN or a subjectAlternativeName.")

        pre_issue_cert.send(sender=self.__class__, ca=ca, csr=csr, expires=expires, algorithm=algorithm,
                            subject=cert_subject, extensions=cert_extensions, password=password)

        public_key = csr.public_key()
        builder = get_cert_builder(expires)
        builder = builder.public_key(public_key)
        builder = builder.issuer_name(issuer_name)
        builder = builder.subject_name(cert_subject.name)

        for key, extension in cert_extensions.items():
            builder = builder.add_extension(*extension.for_builder())

        # Add the SubjectKeyIdentifier
        if SubjectKeyIdentifier.key not in cert_extensions:
            builder = builder.add_extension(x509.SubjectKeyIdentifier.from_public_key(public_key),
                                            critical=False)

        return builder.sign(private_key=ca.key(password), algorithm=algorithm, backend=default_backend())

    def copy(self):
        return deepcopy(self)

    def serialize(self):
        """Function to serialize a profile.

        This is function is called by the admin interface to retrieve profile information to the browser, so
        the value returned by this function should always be JSON serializable.
        """
        data = {
            'cn_in_san': self.cn_in_san,
            'description': self.description,
            'subject': dict(self.subject),
            'extensions': {k: e.serialize() for k, e in self.extensions.items()},
        }

        return data

    def _update_from_ca(self, ca, extensions, add_crl_url=None, add_ocsp_url=None, add_issuer_url=None,
                        add_issuer_alternative_name=None):
        """Update data from the given CA.

        * Sets the AuthorityKeyIdentifier extension
        * Sets the OCSP url if add_ocsp_url is True
        * Sets a CRL URL if add_crl_url is True
        * Adds an IssuerAlternativeName if add_issuer_alternative_name is True

        """
        extensions.setdefault(AuthorityKeyIdentifier.key, ca.get_authority_key_identifier_extension())

        if add_crl_url is not False and ca.crl_url:
            extensions.setdefault(CRLDistributionPoints.key, CRLDistributionPoints())
            extensions[CRLDistributionPoints.key].value.append(DistributionPoint({
                'full_name': [url.strip() for url in ca.crl_url.split()],
            }))

        if add_ocsp_url is not False and ca.ocsp_url:
            extensions.setdefault(AuthorityInformationAccess.key, AuthorityInformationAccess())
            extensions[AuthorityInformationAccess.key].value['ocsp'].append(parse_general_name(ca.ocsp_url))

        if add_issuer_url is not False and ca.issuer_url:
            extensions.setdefault(AuthorityInformationAccess.key, AuthorityInformationAccess())
            extensions[AuthorityInformationAccess.key].value['issuers'].append(
                parse_general_name(ca.issuer_url)
            )
        if add_issuer_alternative_name is not False and ca.issuer_alt_name:
            extensions.setdefault(IssuerAlternativeName.key, IssuerAlternativeName())
            extensions[IssuerAlternativeName.key].extend(shlex_split(ca.issuer_alt_name, ','))

        if self.issuer_name:
            return self.issuer_name.name
        else:
            return ca.x509.subject

    def _update_san_from_cn(self, cn_in_san, subject, extensions):
        if subject.get('CN') and cn_in_san is True:
            try:
                cn = parse_general_name(subject['CN'])
            except ValueError as e:
                raise ValueError(
                    '%s: Could not parse CommonName as subjectAlternativeName.' % subject['CN']) from e

            extensions.setdefault(SubjectAlternativeName.key, SubjectAlternativeName())
            if cn not in extensions[SubjectAlternativeName.key]:
                extensions[SubjectAlternativeName.key].append(cn)
        elif not subject.get('CN') and SubjectAlternativeName.key in extensions:
            cn = extensions[SubjectAlternativeName.key].get_common_name()
            if cn is not None:
                subject['CN'] = cn


def get_profile(name=None):
    """Get profile by the given name.

    Raises ``KeyError`` if the profile is not defined.

    Parameters
    ----------

    name : str, optional
        The name of the profile. If ``None``, the profile configured by
        :ref:`CA_DEFAULT_PROFILE <settings-ca-default-profile>` is used.
    """
    if name is None:
        name = ca_settings.CA_DEFAULT_PROFILE
    return Profile(name, **ca_settings.CA_PROFILES[name])


class Profiles:
    def __init__(self):
        self._profiles = local()

    def __getitem__(self, name):
        if name is None:
            name = ca_settings.CA_DEFAULT_PROFILE

        try:
            return self._profiles.profiles[name]
        except AttributeError:
            self._profiles.profiles = {}
        except KeyError:
            pass

        self._profiles.profiles[name] = get_profile(name)
        return self._profiles.profiles[name]

    def _reset(self):
        self._profiles = local()


profiles = Profiles()


class DefaultProfileProxy:
    """Default profile proxy, similar to Djangos DefaultCacheProxy.

    .. NOTE:: We don't implement setattr/delattr, because Profiles are supposed to be read-only anyway.
    """

    def __getattr__(self, name):
        return getattr(profiles[ca_settings.CA_DEFAULT_PROFILE], name)

    def __eq__(self, other):
        return profiles[ca_settings.CA_DEFAULT_PROFILE] == other

    def __repr__(self):
        return '<DefaultProfile: %r>' % self.name

    def __str__(self):
        return repr(self)


profile = DefaultProfileProxy()



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to create a certificate authority.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

import os
import pathlib
from datetime import timedelta

from django.core.management.base import CommandError
from django.utils import timezone

from ... import ca_settings
from ...extensions import IssuerAlternativeName
from ...extensions import NameConstraints
from ...models import CertificateAuthority
from ...tasks import cache_crl
from ...tasks import generate_ocsp_key
from ...tasks import run_task
from ..base import BaseCommand
from ..base import CertificateAuthorityDetailMixin
from ..base import ExpiresAction
from ..base import MultipleURLAction
from ..base import PasswordAction
from ..base import URLAction


class Command(BaseCommand, CertificateAuthorityDetailMixin):  # pylint: disable=missing-class-docstring
    """Create a certificate authority."""
    help = "Create a certificate authority."

    def add_arguments(self, parser):
        self.add_general_args(parser)
        self.add_algorithm(parser)

        self.add_key_type(parser)
        self.add_key_size(parser)
        self.add_ecc_curve(parser)

        parser.add_argument(
            '--expires', metavar='DAYS', action=ExpiresAction, default=timedelta(365 * 10),
            help='CA certificate expires in DAYS days (default: %(default)s).'
        )
        self.add_ca(
            parser, '--parent', no_default=True,
            help='''Make the CA an intermediate CA of the named CA. By default, this is a new root CA.''')
        parser.add_argument('name', help='Human-readable name of the CA')
        self.add_subject(
            parser, help='''The subject of the CA in the format "/key1=value1/key2=value2/...",
                            valid keys are %s. If "CN" is not set, the name is used.'''
            % self.valid_subject_keys)
        self.add_password(
            parser, help='Optional password used to encrypt the private key. If no argument is passed, you '
                         'will be prompted.')
        parser.add_argument('--path', type=pathlib.PurePath,
                            help="Path where to store Certificate Authorities (relative to CA_DIR).")
        parser.add_argument('--parent-password', nargs='?', action=PasswordAction, metavar='PASSWORD',
                            prompt='Password for parent CA: ',
                            help='Password for the private key of any parent CA.')

        group = parser.add_argument_group(
            'Default hostname',
            'The default hostname is used to compute default URLs for services like OCSP. The hostname is '
            'usually configured in your settings (current setting: %s), but you can override that value '
            'here. The value must be just the hostname and optionally a port, *without* a protocol, e.g. '
            '"ca.example.com" or "ca.example.com:8000".'
            % ca_settings.CA_DEFAULT_HOSTNAME
        )
        group = group.add_mutually_exclusive_group()
        group.add_argument('--default-hostname', metavar='HOSTNAME',
                           help='Override the the default hostname configured in your settings.')
        group.add_argument('--no-default-hostname', dest='default_hostname', action='store_false',
                           help='Disable any default hostname configured in your settings.')

        self.add_acme_group(parser)

        group = parser.add_argument_group(
            'pathlen attribute',
            """Maximum number of CAs that can appear below this one. A pathlen of zero (the default) means it
            can only be used to sign end user certificates and not further CAs.""")
        group = group.add_mutually_exclusive_group()
        group.add_argument('--pathlen', default=0, type=int,
                           help='Maximum number of sublevel CAs (default: %(default)s).')
        group.add_argument('--no-pathlen', action='store_const', const=None, dest='pathlen',
                           help='Do not add a pathlen attribute.')

        group = parser.add_argument_group(
            'X509 v3 certificate extensions for CA',
            '''Extensions added to the certificate authority itself. These options cannot be changed without
            creating a new authority.'''
        )
        group.add_argument(
            '--ca-crl-url', metavar='URL', action=MultipleURLAction, default=[],
            help='URL to a certificate revokation list. Can be given multiple times.'
        )
        group.add_argument(
            '--ca-ocsp-url', metavar='URL', action=URLAction,
            help='URL of an OCSP responder.'
        )
        group.add_argument('--ca-issuer-url', metavar='URL', action=URLAction,
                           help='URL to the certificate of your CA (in DER format).')

        nc_group = parser.add_argument_group(
            'Name Constraints',
            "Add name constraints to the CA, limiting what certificates this CA can sign."
        )
        nc_group.add_argument(
            '--permit-name', metavar='NAME', action='append', default=[],
            help='Add the given name to the permitted-subtree.'
        )
        nc_group.add_argument(
            '--exclude-name', metavar='NAME', action='append', default=[],
            help='Add the given name to the excluded-subtree.'
        )

        self.add_ca_args(parser)

    def handle(self, name, subject, **options):  # pylint: disable=arguments-differ
        if not os.path.exists(ca_settings.CA_DIR):  # pragma: no cover
            # TODO: set permissions
            os.makedirs(ca_settings.CA_DIR)

        # In case of CAs, we silently set the expiry date to that of the parent CA if the user specified a
        # number of days that would make the CA expire after the parent CA.
        #
        # The reasoning is simple: When issuing the child CA, the default is automatically after that of the
        # parent if it wasn't issued on the same day.
        parent = options['parent']
        if parent and timezone.now() + options['expires'] > parent.expires:
            options['expires'] = parent.expires
        if parent and not parent.allows_intermediate_ca:
            raise CommandError("Parent CA cannot create intermediate CA due to pathlen restrictions.")
        if not parent and options['ca_crl_url']:
            raise CommandError("CRLs cannot be used to revoke root CAs.")
        if not parent and options['ca_ocsp_url']:
            raise CommandError("OCSP cannot be used to revoke root CAs.")

        # See if we can work with the private key
        if parent:
            self.test_private_key(parent, options['parent_password'])

        # Set CommonName to name if not set in subject
        if 'CN' not in subject:
            subject['CN'] = name

        name_constraints = NameConstraints({
            'value': {
                'permitted': options['permit_name'],
                'excluded': options['exclude_name'],
            }
        })

        issuer_alternative_name = options[IssuerAlternativeName.key]
        if issuer_alternative_name is None:
            issuer_alternative_name = ''

        kwargs = {}
        for opt in ['path', 'parent', 'default_hostname']:
            if options[opt] is not None:
                kwargs[opt] = options[opt]

        if ca_settings.CA_ENABLE_ACME:  # pragma: no branch; never False because parser throws error already
            # These settings are only there if ACME is enabled
            for opt in ['acme_enabled', 'acme_requires_contact']:
                if options[opt] is not None:
                    kwargs[opt] = options[opt]

        try:
            ca = CertificateAuthority.objects.init(
                key_size=options['key_size'], key_type=options['key_type'],
                ecc_curve=options['ecc_curve'],
                algorithm=options['algorithm'],
                expires=options['expires'],
                pathlen=options['pathlen'],
                issuer_url=options['issuer_url'],
                issuer_alt_name=','.join(issuer_alternative_name),
                crl_url=options['crl_url'],
                ocsp_url=options['ocsp_url'],
                ca_issuer_url=options['ca_issuer_url'],
                ca_crl_url=options['ca_crl_url'],
                ca_ocsp_url=options['ca_ocsp_url'],
                name_constraints=name_constraints,
                name=name, subject=subject, password=options['password'],
                parent_password=options['parent_password'],
                caa=options['caa'],
                website=options['website'],
                terms_of_service=options['tos'],
                **kwargs
            )
        except Exception as ex:
            raise CommandError(ex) from ex

        # Generate OCSP keys and cache CRLs
        run_task(generate_ocsp_key, serial=ca.serial, password=options['password'])
        run_task(cache_crl, serial=ca.serial, password=options['password'])



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to notify watchers about expiring certificates.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from datetime import timedelta

from django.conf import settings
from django.core.mail import send_mail
from django.core.management.base import BaseCommand
from django.utils import timezone

from ... import ca_settings
from ...models import Certificate


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "Send notifications about expiring certificates to watchers."

    def add_arguments(self, parser):
        parser.add_argument('--days', type=int, default=14,
                            help='Warn DAYS days ahead of time (default: %(default)s).')

    def handle(self, *args, **options):  # pylint: disable=arguments-differ
        now = timezone.now()
        expires = now + timedelta(days=options['days'] + 1)  # add a day to avoid one-of errors

        qs = Certificate.objects.valid().filter(expires__lt=expires)
        for cert in qs:
            days = (cert.expires - now).days

            if days not in ca_settings.CA_NOTIFICATION_DAYS:
                continue

            timestamp = cert.expires.strftime('%Y-%m-%d')
            subj = 'Certificate expiration for %s on %s' % (cert.cn, timestamp)
            msg = 'The certificate for %s will expire on %s.' % (cert.cn, timestamp)
            recipient = list(cert.watchers.values_list('mail', flat=True))
            send_mail(subj, msg, settings.DEFAULT_FROM_EMAIL, recipient)



# Generated by Django 2.2 on 2019-04-27 16:26

from django.db import migrations, models
import django_ca.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0012_auto_20190405_2345'),
    ]

    operations = [
        migrations.AddField(
            model_name='certificateauthority',
            name='crl_number',
            field=models.TextField(blank=True, default='{"scope": {}}', help_text='Data structure to store the CRL number (see RFC 5280, 5.2.3) depending on the scope.', validators=[django_ca.models.json_validator], verbose_name='CRL Number'),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Keep track of internal settings for django-ca."""

import os
import re
from datetime import timedelta

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding

from django.conf import global_settings
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _

if 'CA_DIR' in os.environ:  # pragma: no cover
    CA_DIR = os.path.join(os.environ['CA_DIR'], 'files')
else:
    CA_DIR = getattr(settings, 'CA_DIR', os.path.join(settings.BASE_DIR, 'files'))

CA_DEFAULT_KEY_SIZE = getattr(settings, 'CA_DEFAULT_KEY_SIZE', 4096)

CA_PROFILES = {
    'client': {
        # see: http://security.stackexchange.com/questions/68491/
        'description': _('A certificate for a client.'),
        'extensions': {
            'key_usage': {
                'critical': True,
                'value': [
                    'digitalSignature',
                ],
            },
            'extended_key_usage': {
                'critical': False,
                'value': [
                    'clientAuth',
                ],
            },
        },
    },
    'server': {
        'description': _('A certificate for a server, allows client and server authentication.'),
        'extensions': {
            'key_usage': {
                'critical': True,
                'value': [
                    'digitalSignature',
                    'keyAgreement',
                    'keyEncipherment',
                ],
            },
            'extended_key_usage': {
                'critical': False,
                'value': [
                    'clientAuth',
                    'serverAuth',
                ],
            },
        }
    },
    'webserver': {
        # see http://security.stackexchange.com/questions/24106/
        'description': _('A certificate for a webserver.'),
        'extensions': {
            'key_usage': {
                'critical': True,
                'value': [
                    'digitalSignature',
                    'keyAgreement',
                    'keyEncipherment',
                ],
            },
            'extended_key_usage': {
                'critical': False,
                'value': [
                    'serverAuth',
                ],
            },
        }
    },
    'enduser': {
        # see: http://security.stackexchange.com/questions/30066/
        'description': _(
            'A certificate for an enduser, allows client authentication, code and email signing.'),
        'cn_in_san': False,
        'extensions': {
            'key_usage': {
                'critical': True,
                'value': [
                    'dataEncipherment',
                    'digitalSignature',
                    'keyEncipherment',
                ],
            },
            'extended_key_usage': {
                'critical': False,
                'value': [
                    'clientAuth',
                    'codeSigning',
                    'emailProtection',
                ],
            },
        },
    },
    'ocsp': {
        'description': _('A certificate for an OCSP responder.'),
        'cn_in_san': False,  # CAs frequently use human readable name as CN
        'add_ocsp_url': False,
        'autogenerated': True,
        'extensions': {
            'key_usage': {
                'value': [
                    'nonRepudiation',
                    'digitalSignature',
                    'keyEncipherment',
                ],
            },
            'extended_key_usage': {
                'value': [
                    'OCSPSigning',
                ],
            },
        }
    },
}

_CA_CRL_PROFILES = {
    'user': {
        'algorithm': 'SHA512',
        'expires': 86400,
        'scope': 'user',
        'encodings': ['PEM', 'DER', ],
    },
    'ca': {
        'algorithm': 'SHA512',
        'expires': 86400,
        'scope': 'ca',
        'encodings': ['PEM', 'DER', ],
    }
}

# Get and sanitize default CA serial
# NOTE: This effectively duplicates utils.sanitize_serial()
CA_DEFAULT_CA = getattr(settings, 'CA_DEFAULT_CA', '').replace(':', '').upper()
if CA_DEFAULT_CA != '0':
    CA_DEFAULT_CA = CA_DEFAULT_CA.lstrip('0')
if re.search('[^0-9A-F]', CA_DEFAULT_CA):
    raise ImproperlyConfigured('CA_DEFAULT_CA: %s: Serial contains invalid characters.' % CA_DEFAULT_CA)

CA_DEFAULT_SUBJECT = getattr(settings, 'CA_DEFAULT_SUBJECT', {})

# Add ability just override/add some profiles
_CA_PROFILE_OVERRIDES = getattr(settings, 'CA_PROFILES', {})
for name, profile in _CA_PROFILE_OVERRIDES.items():
    if profile is None:
        del CA_PROFILES[name]

    elif name in CA_PROFILES:
        CA_PROFILES[name].update(profile)
    else:
        CA_PROFILES[name] = profile

for name, profile in CA_PROFILES.items():
    profile.setdefault('subject', CA_DEFAULT_SUBJECT)
    profile.setdefault('cn_in_san', True)

CA_DEFAULT_ENCODING = getattr(settings, 'CA_DEFAULT_ENCODING', Encoding.PEM)
CA_DEFAULT_EXPIRES = getattr(settings, 'CA_DEFAULT_EXPIRES', 730)
CA_DEFAULT_PROFILE = getattr(settings, 'CA_DEFAULT_PROFILE', 'webserver')
CA_NOTIFICATION_DAYS = getattr(settings, 'CA_NOTIFICATION_DAYS', [14, 7, 3, 1, ])
CA_CRL_PROFILES = getattr(settings, 'CA_CRL_PROFILES', _CA_CRL_PROFILES)
CA_PASSWORDS = getattr(settings, 'CA_PASSWORDS', {})

# ACME settings
ACME_ORDER_VALIDITY = getattr(settings, 'CA_ACME_ORDER_VALIDITY', timedelta(hours=1))
ACME_ACCOUNT_REQUIRES_CONTACT = getattr(settings, 'CA_ACME_ACCOUNT_REQUIRES_CONTACT', True)
ACME_MAX_CERT_VALIDITY = getattr(settings, 'CA_ACME_MAX_CERT_VALIDITY', timedelta(days=90))
ACME_DEFAULT_CERT_VALIDITY = getattr(settings, 'CA_ACME_DEFAULT_CERT_VALIDITY', timedelta(days=90))

# Undocumented options, e.g. to share values between different parts of code
CA_MIN_KEY_SIZE = getattr(settings, 'CA_MIN_KEY_SIZE', 2048)

CA_DEFAULT_HOSTNAME = getattr(settings, 'CA_DEFAULT_HOSTNAME', None)

CA_DIGEST_ALGORITHM = getattr(settings, 'CA_DIGEST_ALGORITHM', "sha512").strip().upper()
try:
    CA_DIGEST_ALGORITHM = getattr(hashes, CA_DIGEST_ALGORITHM)()
except AttributeError:
    # pylint: disable=raise-missing-from; not really useful in this context
    raise ImproperlyConfigured('Unkown CA_DIGEST_ALGORITHM: %s' % settings.CA_DIGEST_ALGORITHM)

if isinstance(CA_DEFAULT_EXPIRES, int):
    CA_DEFAULT_EXPIRES = timedelta(days=CA_DEFAULT_EXPIRES)
if isinstance(ACME_MAX_CERT_VALIDITY, int):
    ACME_MAX_CERT_VALIDITY = timedelta(days=ACME_MAX_CERT_VALIDITY)
if isinstance(ACME_DEFAULT_CERT_VALIDITY, int):
    ACME_DEFAULT_CERT_VALIDITY = timedelta(days=ACME_DEFAULT_CERT_VALIDITY)
if isinstance(ACME_ORDER_VALIDITY, int):
    ACME_ORDER_VALIDITY = timedelta(days=ACME_ORDER_VALIDITY)
elif not isinstance(CA_DEFAULT_EXPIRES, timedelta):
    raise ImproperlyConfigured('CA_DEFAULT_EXPIRES: %s: Must be int or timedelta' % CA_DEFAULT_EXPIRES)
if CA_DEFAULT_EXPIRES <= timedelta():
    raise ImproperlyConfigured('CA_DEFAULT_EXPIRES: %s: Must have positive value' % CA_DEFAULT_EXPIRES)

if CA_MIN_KEY_SIZE > CA_DEFAULT_KEY_SIZE:
    raise ImproperlyConfigured('CA_DEFAULT_KEY_SIZE cannot be lower then %s' % CA_MIN_KEY_SIZE)

_CA_DEFAULT_ECC_CURVE = getattr(settings, 'CA_DEFAULT_ECC_CURVE', 'SECP256R1').strip()
try:
    CA_DEFAULT_ECC_CURVE = getattr(ec, _CA_DEFAULT_ECC_CURVE)()
    if not isinstance(CA_DEFAULT_ECC_CURVE, ec.EllipticCurve):
        raise ImproperlyConfigured('%s: Not an EllipticCurve.' % _CA_DEFAULT_ECC_CURVE)
except AttributeError:
    # pylint: disable=raise-missing-from; not really useful in this context
    raise ImproperlyConfigured('Unkown CA_DEFAULT_ECC_CURVE: %s' % settings.CA_DEFAULT_ECC_CURVE)

CA_FILE_STORAGE = getattr(settings, 'CA_FILE_STORAGE', global_settings.DEFAULT_FILE_STORAGE)
CA_FILE_STORAGE_KWARGS = getattr(settings, 'CA_FILE_STORAGE_KWARGS', {
    'location': CA_DIR,
    'file_permissions_mode': 0o600,
    'directory_permissions_mode': 0o700,
})

CA_FILE_STORAGE_URL = 'https://django-ca.readthedocs.io/en/latest/update.html#update-to-1-12-0-or-later'

CA_ENABLE_ACME = getattr(settings, 'CA_ENABLE_ACME', False)

# Decide if we should use Celery or not
CA_USE_CELERY = getattr(settings, 'CA_USE_CELERY', None)
if CA_USE_CELERY is None:
    try:
        from celery import shared_task  # pylint: disable=unused-import
        CA_USE_CELERY = True
    except ImportError:
        CA_USE_CELERY = False
elif CA_USE_CELERY is True:
    try:
        from celery import shared_task  # NOQA: F401
    except ImportError:
        # pylint: disable=raise-missing-from; not really useful in this context
        raise ImproperlyConfigured('CA_USE_CELERY set to True, but Celery is not installed')



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""ACME utility functions."""

import josepy as jose

from cryptography import x509
from cryptography.hazmat.backends import default_backend


def parse_acme_csr(value):
    """Convert the CSR as received via ACMEv2 into a valid CSR.

    ACMEv2 sends the CSR as a base64url encoded string of its DER /ASN.1 representation.

    Returns
    -------

    :py:class:`~cg:cryptography.x509.CertificateSigningRequest`
        The CSR as used by cryptography.
    """
    decoded = jose.decode_b64jose(value)
    return x509.load_der_x509_csr(decoded, default_backend())



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

import copy
import json
import logging
from datetime import datetime
from functools import partial
from types import MethodType

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import Encoding

from django.contrib import admin
from django.contrib.messages import constants as messages
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseRedirect
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.urls import path
from django.urls import reverse
from django.utils import timezone
from django.utils.encoding import force_bytes
from django.utils.html import escape
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from django_object_actions import DjangoObjectActions

from . import ca_settings
from .constants import ReasonFlags
from .extensions import KEY_TO_EXTENSION
from .extensions import AlternativeNameExtension
from .extensions import CRLDistributionPointsBase
from .extensions import ExtendedKeyUsage
from .extensions import KeyUsage
from .extensions import NullExtension
from .extensions import OrderedSetExtension
from .extensions import SubjectAlternativeName
from .extensions import TLSFeature
from .extensions import UnrecognizedExtension
from .forms import CreateCertificateForm
from .forms import ResignCertificateForm
from .forms import RevokeCertificateForm
from .forms import X509CertMixinAdminForm
from .models import AcmeAccount
from .models import AcmeAuthorization
from .models import AcmeCertificate
from .models import AcmeChallenge
from .models import AcmeOrder
from .models import Certificate
from .models import CertificateAuthority
from .models import Watcher
from .profiles import profiles
from .signals import post_issue_cert
from .utils import OID_NAME_MAPPINGS
from .utils import SERIAL_RE
from .utils import LazyEncoder
from .utils import add_colons

log = logging.getLogger(__name__)


@admin.register(Watcher)
class WatcherAdmin(admin.ModelAdmin):
    pass


class CertificateMixin(object):
    form = X509CertMixinAdminForm

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        urls = [
            path('<int:pk>/download/', self.admin_site.admin_view(self.download_view),
                 name='%s_%s_download' % info),
            path('<int:pk>/download_bundle/', self.admin_site.admin_view(self.download_bundle_view),
                 name='%s_%s_download_bundle' % info),
        ]
        urls += super(CertificateMixin, self).get_urls()
        return urls

    def _download_response(self, request, pk, bundle=False):
        if not request.user.is_staff or not self.has_change_permission(request):
            # NOTE: is_staff is already assured by ModelAdmin, but just to be sure
            raise PermissionDenied

        # get object in question
        try:
            obj = self.model.objects.get(pk=pk)
        except self.model.DoesNotExist:
            raise Http404

        # get filetype
        filetype = request.GET.get('format', 'PEM').upper().strip()

        if filetype == 'PEM':
            if bundle is True:
                data = '\n'.join([cert.pub.strip() for cert in obj.bundle])
            else:
                data = obj.pub
        elif filetype == 'DER':
            if bundle is True:
                return HttpResponseBadRequest(_('DER/ASN.1 certificates cannot be downloaded as a bundle.'))
            data = obj.x509.public_bytes(encoding=Encoding.DER)
        else:
            return HttpResponseBadRequest()

        filename = obj.get_filename(ext=filetype, bundle=bundle)
        response = HttpResponse(data, content_type='application/pkix-cert')
        response['Content-Disposition'] = 'attachment; filename=%s' % filename
        return response

    def download_view(self, request, pk):
        """A view that allows the user to download a certificate in PEM or DER/ASN1 format."""

        return self._download_response(request, pk)

    def download_bundle_view(self, request, pk):
        """A view that allows the user to download a certificate bundle in PEM format."""

        return self._download_response(request, pk, bundle=True)

    def has_delete_permission(self, request, obj=None):
        return False

    def get_actions(self, request):
        """Disable the "delete selected" admin action.

        Otherwise the action is present even though has_delete_permission is False, it just doesn't
        work.
        """
        actions = super(CertificateMixin, self).get_actions(request)
        actions.pop('delete_selected', '')
        return actions

    def hpkp_pin(self, obj):
        return obj.hpkp_pin
    hpkp_pin.short_description = _('HPKP pin')

    def cn_display(self, obj):
        if obj.cn:
            return obj.cn
        return _('<none>')
    cn_display.short_description = _('CommonName')

    def serial_field(self, obj):
        return add_colons(obj.serial)
    serial_field.short_description = _('Serial')
    serial_field.admin_order_field = 'serial'

    def get_search_results(self, request, queryset, search_term):
        # Replace ':' from any search term that looks like a serial
        search_term = ' '.join([
            t.replace(':', '').upper() if SERIAL_RE.match(t.upper().strip(':')) else t
            for t in search_term.split()
        ])

        return super(CertificateMixin, self).get_search_results(request, queryset, search_term)

    ##################################
    # Properties for x509 extensions #
    ##################################

    def output_template(self, obj, key):
        ext = getattr(obj, key)
        templates = ['django_ca/admin/extensions/%s.html' % key]

        if isinstance(ext, NullExtension):
            templates.append('django_ca/admin/extensions/base/null_extension.html')
        if isinstance(ext, AlternativeNameExtension):
            templates.append('django_ca/admin/extensions/base/alternative_name_extension.html')
        if isinstance(ext, CRLDistributionPointsBase):
            templates.append('django_ca/admin/extensions/base/crl_distribution_points_base.html')
        if isinstance(ext, OrderedSetExtension):
            templates.append('django_ca/admin/extensions/base/ordered_set_extension.html')
        if isinstance(ext, UnrecognizedExtension) \
                or isinstance(ext, x509.UnrecognizedExtension):  # pragma: no cover
            templates.append('django_ca/admin/extensions/base/unrecognized_extension.html')
        else:
            templates.append('django_ca/admin/extensions/base/base.html')
        return render_to_string(templates, {'obj': obj, 'extension': ext})

    def unknown_oid(self, oid, obj):
        ext = obj.x509.extensions.get_extension_for_oid(oid)
        html = ''
        if ext.critical is True:
            text = _('Critical')
            html = '<img src="/static/admin/img/icon-yes.svg" alt="%s"> %s' % (text, text)

        html += '<p>%s<p>' % escape(ext.value)
        return html

    def get_oid_name(self, oid):
        return oid.dotted_string.replace('.', '_')

    def get_fieldsets(self, request, obj=None):
        fieldsets = super(CertificateMixin, self).get_fieldsets(request, obj=obj)

        if obj is None:
            return fieldsets

        fieldsets = copy.deepcopy(fieldsets)

        if obj.extension_fields:
            for field in obj.extension_fields:
                if field == SubjectAlternativeName.key:  # already displayed in main section
                    continue

                # If we encounter an object of type x509.Extension, it means that we do not yet support this
                # extension, hence there are no accessors either. We compute a name for the extension based on
                # the OID, create a partial function of unknown_oid and attach it under that name to this
                # admin instance:
                if isinstance(field, x509.Extension):
                    func = partial(self.unknown_oid, field.oid)
                    func.short_description = 'Unkown OID (%s)' % field.oid.dotted_string

                    field = self.get_oid_name(field.oid)

                    # attach function to this instance
                    setattr(self, field, func)

                fieldsets[self.x509_fieldset_index][1]['fields'].append(field)

        else:
            # we have no extensions, so remove the whole fieldset
            fieldsets.pop(self.x509_fieldset_index)

        return fieldsets

    def get_readonly_fields(self, request, obj=None):
        fields = super(CertificateMixin, self).get_readonly_fields(request, obj=obj)

        if not obj.revoked:
            # We can only change the date when the certificate was compromised if it's actually revoked.
            fields.append('compromised')

        if obj is None:  # pragma: no cover
            # This is never True because CertificateAdmin (the only case where objects are added) doesn't call
            # the superclass in this case.
            return fields

        fields = list(fields)
        for field in obj.extension_fields:
            if isinstance(field, x509.Extension):
                field = self.get_oid_name(field.oid)

            fields.append(field)

        return fields

    class Media:
        css = {
            'all': (
                'django_ca/admin/css/base.css',
            ),
        }


# Attach extension properties to admin if they are not already present.
# This makes ModelAdmin have a property for every extension that we currently support,
# rendering as a template based on the extension key.
for key, ext in KEY_TO_EXTENSION.items():
    # Give Mixin opportunity to override method - not needed right now
    #if hasattr(CertificateMixin, key):
    #    continue

    f = partial(CertificateMixin.output_template, key=key)
    f.short_description = ext.name
    f = MethodType(f, CertificateMixin)

    setattr(CertificateMixin, key, f)


@admin.register(CertificateAuthority)
class CertificateAuthorityAdmin(CertificateMixin, admin.ModelAdmin):
    fieldsets = (
        (None, {
            'fields': ['name', 'enabled', 'cn_display', 'parent', 'hpkp_pin', 'caa_identity', 'website',
                       'terms_of_service'],
        }),
        (_('Details'), {
            'description': _('Information to add to newly signed certificates.'),
            'fields': ['crl_url', 'crl_number', 'issuer_url', 'ocsp_url', 'issuer_alt_name', ],
        }),
        (_('Certificate'), {
            'fields': ['serial_field', 'pub', 'expires'],
            # The "as-code" class is used so CSS can only match this section (and only in an
            # existing cert).
            'classes': ('as-code', ),
        }),
        (_('X509 extensions'), {
            'fields': [],  # dynamically added by add_fieldsets
        }),
    )
    list_display = ['enabled', 'name', 'serial_field', ]
    list_display_links = ['enabled', 'name', ]
    search_fields = ['cn', 'name', 'serial_field', ]
    readonly_fields = ['serial_field', 'pub', 'parent', 'cn_display', 'expires', 'hpkp_pin', ]
    x509_fieldset_index = 3

    def has_add_permission(self, request):
        return False

    def get_fieldsets(self, request, obj=None):
        """Collapse the "Revocation" section unless the certificate is revoked."""
        fieldsets = super().get_fieldsets(request, obj=obj)
        if ca_settings.CA_ENABLE_ACME:
            fieldsets = list(copy.deepcopy(fieldsets))
            fieldsets.insert(1, (_('ACME'), {
                'fields': ['acme_enabled', 'acme_requires_contact', ],
            }))

        return fieldsets

    class Media:
        css = {
            'all': (
                'django_ca/admin/css/base.css',
                'django_ca/admin/css/certificateauthorityadmin.css',
            ),
        }


class DefaultListFilter(admin.SimpleListFilter):
    """Filter that lists only manually created certs by default.

    Inspired by https://stackoverflow.com/a/16556771.
    """

    def choices(self, cl):
        for lookup, title in self.lookup_choices:
            yield {
                'selected': self.value() == lookup,
                'query_string': cl.get_query_string({
                    self.parameter_name: lookup,
                }, []),
                'display': title,
            }


class StatusListFilter(DefaultListFilter):
    title = _('Status')
    parameter_name = 'status'

    def lookups(self, request, model_admin):
        return (
            (None, _('Valid')),
            ('expired', _('Expired')),
            ('revoked', _('Revoked')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        if self.value() is None:
            return queryset.valid()
        elif self.value() == 'expired':
            return queryset.expired()
        elif self.value() == 'revoked':
            return queryset.revoked()


class AutoGeneratedFilter(DefaultListFilter):
    title = _('autogeneration')
    parameter_name = 'auto'

    def lookups(self, request, model_admin):
        return (
            (None, _('No')),
            ('auto', _('Yes')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        if self.value() == 'auto':
            return Certificate.objects.filter(autogenerated=True)
        elif self.value() is None:
            return Certificate.objects.filter(autogenerated=False)
        # "all" does not need a filter


@admin.register(Certificate)
class CertificateAdmin(DjangoObjectActions, CertificateMixin, admin.ModelAdmin):
    actions = ['revoke', ]
    change_actions = ('revoke_change', 'resign', )
    add_form_template = 'admin/django_ca/certificate/add_form.html'
    change_form_template = 'admin/django_ca/certificate/change_form.html'
    list_display = ('cn_display', 'profile', 'serial_field', 'status', 'expires_date')
    list_filter = ('profile', AutoGeneratedFilter, StatusListFilter, 'ca')
    readonly_fields = [
        'expires', 'csr', 'pub', 'cn_display', 'serial_field', 'revoked', 'revoked_date', 'revoked_reason',
        'distinguished_name', 'ca', 'hpkp_pin', 'subject_alternative_name', 'profile', ]
    search_fields = ['cn', 'serial', ]

    fieldsets = [
        (None, {
            'fields': ['cn_display', 'subject_alternative_name', 'distinguished_name', 'serial_field', 'ca',
                       ('expires', 'autogenerated'), 'watchers', 'hpkp_pin', 'profile', ],
        }),
        (_('X.509 Extensions'), {
            'fields': [],
            'classes': ('collapse', ),
        }),
        (_('Revocation'), {
            'fields': (('revoked', 'revoked_reason'),
                       ('revoked_date', 'compromised', )),
        }),
        (_('Certificate'), {
            'fields': ['pub', 'csr', ],
            # The "as-code" class is used so CSS can only match this section (and only in an
            # existing cert).
            'classes': ('collapse', 'as-code'),
        }),
    ]
    add_fieldsets = [
        (None, {
            'fields': ['csr', ('ca', 'password'), 'profile', 'subject', 'subject_alternative_name',
                       'algorithm', ('expires', 'autogenerated'), 'watchers', ],
        }),
        (_('X.509 Extensions'), {
            'fields': ['key_usage', 'extended_key_usage', 'tls_feature', ]
        }),
    ]

    # same as add_fieldsets but without the csr
    resign_fieldsets = [
        (None, {
            'fields': [('ca', 'password'), 'profile', 'subject', 'subject_alternative_name', 'algorithm',
                       'expires', 'watchers', ],
        }),
        (_('X.509 Extensions'), {
            'fields': ['key_usage', 'extended_key_usage', 'tls_feature', ]
        }),
    ]
    x509_fieldset_index = 1

    def has_add_permission(self, request):
        # Only grant add permissions if there is at least one useable CA
        for ca in CertificateAuthority.objects.filter(enabled=True):
            if ca.key_exists:
                return True
        return False

    def get_form(self, request, obj=None, **kwargs):
        if hasattr(request, '_resign_obj'):
            return ResignCertificateForm
        elif obj is None:
            return CreateCertificateForm
        else:
            return super(CertificateAdmin, self).get_form(request, obj=obj, **kwargs)

    def get_changeform_initial_data(self, request):
        data = super(CertificateAdmin, self).get_changeform_initial_data(request)

        if hasattr(request, '_resign_obj'):
            # resign the cert, so we add initial data from the original cert

            resign_obj = getattr(request, '_resign_obj')
            san = resign_obj.subject_alternative_name
            if san is None:
                san = ('', False)
            else:
                san = (','.join(san), False)
            algo = resign_obj.algorithm.__class__.__name__

            data = {
                'algorithm': algo,
                'ca': resign_obj.ca,
                'extended_key_usage': resign_obj.extended_key_usage,
                'key_usage': resign_obj.key_usage,
                # TODO: pass profile once it's stored to the database
                #'profile': '',
                'subject': resign_obj.subject,
                'subject_alternative_name': san,
                'tls_feature': resign_obj.tls_feature,
                'watchers': resign_obj.watchers.all(),
            }
        else:
            data['subject'] = ca_settings.CA_PROFILES[ca_settings.CA_DEFAULT_PROFILE].get('subject', {})

        return data

    def add_view(self, request, form_url='', extra_context=None):
        extra_context = extra_context or {}
        extra_context['profiles_url'] = reverse('admin:%s' % self.profiles_view_name)
        extra_context['csr_details_url'] = reverse('admin:%s' % self.csr_details_view_name)
        return super(CertificateAdmin, self).add_view(request, form_url=form_url, extra_context=extra_context)

    @property
    def csr_details_view_name(self):
        return '%s_%s_csr_details' % (self.model._meta.app_label, self.model._meta.verbose_name)

    def csr_details_view(self, request):
        """Returns details of a CSR request."""

        if not request.user.is_staff or not self.has_change_permission(request):
            # NOTE: is_staff is already assured by ModelAdmin, but just to be sure
            raise PermissionDenied

        try:
            csr = x509.load_pem_x509_csr(force_bytes(request.POST['csr']), default_backend())
        except Exception as e:
            return HttpResponseBadRequest(json.dumps({
                'message': str(e),
            }), content_type='application/json')

        subject = {OID_NAME_MAPPINGS[s.oid]: s.value for s in csr.subject}
        return HttpResponse(json.dumps({
            'subject': subject,
        }), content_type='application/json')

    @property
    def profiles_view_name(self):
        return '%s_%s_profiles' % (self.model._meta.app_label, self.model._meta.verbose_name)

    def profiles_view(self, request):
        """Returns profiles."""

        if not request.user.is_staff or not self.has_change_permission(request):
            # NOTE: is_staff is already assured by ModelAdmin, but just to be sure
            raise PermissionDenied

        data = {name: profiles[name].serialize() for name in ca_settings.CA_PROFILES}
        return HttpResponse(json.dumps(data, cls=LazyEncoder), content_type='application/json')

    def get_urls(self):
        # Remove the delete action from the URLs
        urls = super(CertificateAdmin, self).get_urls()

        # add csr-details and profiles
        urls.insert(0, path('ajax/csr-details', self.admin_site.admin_view(self.csr_details_view),
                            name=self.csr_details_view_name))
        urls.insert(0, path('ajax/profiles', self.admin_site.admin_view(self.profiles_view),
                            name=self.profiles_view_name))

        return urls

    def resign(self, request, obj):
        if not obj.csr:
            self.message_user(request, _('Certificate has no CSR (most likely because it was imported).'),
                              messages.ERROR)
            return HttpResponseRedirect(obj.admin_change_url)

        request._resign_obj = obj
        extra_context = {
            'title': _('Resign %s for %s') % (obj._meta.verbose_name, obj),
            'original_obj': obj,
            'object_action': _('Resign'),
        }
        return self.changeform_view(request, extra_context=extra_context)
    resign.short_description = _('Resign this certificate.')

    def revoke_change(self, request, obj):
        if not self.has_change_permission(request, obj):
            raise PermissionDenied
        if obj.revoked:
            self.message_user(request, _('Certificate is already revoked.'), level=messages.ERROR)
            return HttpResponseRedirect(obj.admin_change_url)

        if request.method == 'POST':
            form = RevokeCertificateForm(request.POST, instance=obj)
            if form.is_valid():
                reason = form.cleaned_data['revoked_reason']
                if reason:
                    reason = ReasonFlags[reason]

                obj.revoke(
                    reason=reason,
                    compromised=form.cleaned_data['compromised'] or None
                )
                return HttpResponseRedirect(obj.admin_change_url)
        else:
            form = RevokeCertificateForm(instance=obj)

        context = dict(self.admin_site.each_context(request), form=form, object=obj, opts=obj._meta)
        return TemplateResponse(request, "admin/django_ca/certificate/revoke_form.html", context)
    revoke_change.label = _('Revoke')
    revoke_change.short_description = _('Revoke this certificate')

    def revoke(self, request, queryset):
        for cert in queryset:
            cert.revoke()
    revoke.short_description = _('Revoke selected certificates')

    def get_change_actions(self, request, object_id, form_url):
        actions = list(super(CertificateAdmin, self).get_change_actions(request, object_id, form_url))
        try:
            obj = self.model.objects.get(pk=object_id)
        except self.model.DoesNotExist:
            return []

        if obj.revoked:
            actions.remove('revoke_change')
        return actions

    def get_fieldsets(self, request, obj=None):
        """Collapse the "Revocation" section unless the certificate is revoked."""
        fieldsets = super(CertificateAdmin, self).get_fieldsets(request, obj=obj)

        if hasattr(request, '_resign_obj'):
            return self.resign_fieldsets
        if obj is None:
            return self.add_fieldsets

        if obj.revoked is False:
            fieldsets[2][1]['classes'] = ['collapse', ]
        return fieldsets

    def get_readonly_fields(self, request, obj=None):
        if obj is None:
            return []
        return super(CertificateAdmin, self).get_readonly_fields(request, obj=obj)

    def status(self, obj):
        if obj.revoked:
            return _('Revoked')
        if obj.expires < timezone.now():
            return _('Expired')
        else:
            return _('Valid')
    status.short_description = _('Status')

    def expires_date(self, obj):
        return obj.expires.date()
    expires_date.short_description = _('Expires')
    expires_date.admin_order_field = 'expires'

    def save_model(self, request, obj, form, change):
        data = form.cleaned_data

        # If this is a new certificate, initialize it.
        if change is False:
            profile = profiles[data['profile']]

            if hasattr(request, '_resign_obj'):
                csr = getattr(request, '_resign_obj').csr
                obj.csr = csr
            else:
                # Note: CSR is set by model form already
                csr = data['csr']

            parsed_csr = Certificate.objects.parse_csr(csr, Encoding.PEM)
            expires = datetime.combine(data['expires'], datetime.min.time())

            extensions = {}
            san, cn_in_san = data['subject_alternative_name']
            san = SubjectAlternativeName({'value': [e.strip() for e in san.split(',') if e.strip()]})
            if san:
                extensions[SubjectAlternativeName.key] = san
            for ext_key in [KeyUsage.key, ExtendedKeyUsage.key, TLSFeature.key]:
                if data[ext_key].value:
                    extensions[ext_key] = data[ext_key]
                else:
                    extensions[ext_key] = None

            obj.profile = profile.name
            obj.x509 = profile.create_cert(data['ca'], parsed_csr, subject=data['subject'], expires=expires,
                                           algorithm=data['algorithm'], cn_in_san=cn_in_san,
                                           password=data['password'], extensions=extensions)
            obj.save()
            post_issue_cert.send(sender=self.model, cert=obj)
        else:
            obj.save()

    class Media:
        css = {
            'all': (
                'django_ca/admin/css/base.css',
                'django_ca/admin/css/certificateadmin.css',
            ),
        }
        js = (
            'admin/js/jquery.init.js',
            'django_ca/admin/js/sign.js',
        )


if ca_settings.CA_ENABLE_ACME:  # pragma: no branch
    class ExpiredListFilter(DefaultListFilter):
        """Filter for expired ACME orders."""
        title = _('Expired')
        parameter_name = 'expired'

        def lookups(self, request, model_admin):
            return (
                ('0', _('No')),
                ('1', _('Yes')),
            )

        def queryset(self, request, queryset):
            now = timezone.now()

            if self.value() == '0':
                return queryset.filter(expires__gt=now)
            if self.value() == '1':
                return queryset.filter(expires__lt=now)
            return queryset

    @admin.register(AcmeAccount)
    class AcmeAccountAdmin(admin.ModelAdmin):
        """ModelAdmin class for :py:class:`~django_ca.models.AcmeAccount`."""
        list_display = ('first_contact', 'ca', 'slug', 'status', 'created', 'terms_of_service_agreed')
        list_filter = ('ca', 'status', 'terms_of_service_agreed')
        readonly_fields = ('pem', 'created', )
        search_fields = ('contact', )

        def first_contact(self, obj):
            return str(obj)
        first_contact.short_description = _('Contact')
        first_contact.admin_order_field = 'contact'

    @admin.register(AcmeOrder)
    class AcmeOrderAdmin(admin.ModelAdmin):
        """ModelAdmin class for :py:class:`~django_ca.models.AcmeOrder`."""
        list_display = ('slug', 'ca', 'status', 'account_link', 'expires', )
        list_filter = ('status', ExpiredListFilter)
        list_select_related = ('account', )
        search_fields = ('account__contact', 'slug')

        def ca(self, obj):
            return format_html('<a href="{}">{}</a>', obj.account.ca.admin_change_url, obj.account.ca)
        ca.short_description = _('CA')
        ca.admin_order_field = 'account__ca'

        def account_link(self, obj):
            return format_html('<a href="{}">{}</a>', obj.account.admin_change_url, obj.account)
        account_link.short_description = _('Account')
        account_link.admin_order_field = 'account__contact'

    @admin.register(AcmeAuthorization)
    class AcmeAuthorizationAdmin(admin.ModelAdmin):
        """ModelAdmin class for :py:class:`~django_ca.models.AcmeAuthorization`."""
        list_display = ('slug', 'value', 'status', 'ca', 'account', 'order_display')
        list_filter = ('status', 'order__account__ca', )
        list_select_related = ('order__account__ca', )
        search_fields = ('value', 'slug', 'order__account__contact', )

        def account(self, obj):
            return format_html('<a href="{}">{}</a>', obj.order.account.admin_change_url, obj.order.account)
        account.short_description = _('Account')
        account.admin_order_field = 'order__account__contact'

        def ca(self, obj):
            return format_html('<a href="{}">{}</a>', obj.order.account.ca.admin_change_url,
                               obj.order.account.ca)
        ca.short_description = _('CA')
        ca.admin_order_field = 'account__ca'

        def order_display(self, obj):
            return format_html('<a href="{}">{}</a>', obj.order.admin_change_url, obj.order.slug)
        order_display.short_description = _('Order')
        order_display.admin_order_field = 'order__slug'

    @admin.register(AcmeChallenge)
    class AcmeChallengeAdmin(admin.ModelAdmin):
        """ModelAdmin class for :py:class:`~django_ca.models.AcmeChallenge`."""
        list_display = ('slug', 'auth', 'type', 'status', 'validated', )
        list_filter = ('type', 'status', 'auth__order')

    @admin.register(AcmeCertificate)
    class AcmeCertificateAdmin(admin.ModelAdmin):
        """ModelAdmin class for :py:class:`~django_ca.models.AcmeCertificate`."""

        list_display = ('slug', 'status', 'cert', 'ca', 'account', 'order_link')
        list_filter = ('order__status', 'order__account__ca', )
        list_select_related = ('order__account__ca', )

        def account(self, obj):
            return format_html('<a href="{}">{}</a>', obj.order.account.admin_change_url, obj.order.account)
        account.short_description = _('Account')
        account.admin_order_field = 'order__account__contact'

        def ca(self, obj):
            return format_html('<a href="{}">{}</a>', obj.order.account.ca.admin_change_url,
                               obj.order.account.ca)
        ca.short_description = _('CA')
        ca.admin_order_field = 'order__account__ca'

        def order_link(self, obj):
            return format_html('<a href="{}">{}</a>', obj.order.admin_change_url, obj.order.slug)
        order_link.short_description = _('Order')
        order_link.admin_order_field = 'order'

        def status(self, obj):
            return obj.order.status
        status.short_description = _('Status')
        status.admin_order_field = 'order__status'



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Views for the django-ca app.

.. seealso::

   * https://docs.djangoproject.com/en/dev/topics/class-based-views/
   * https://django-ca.readthedocs.io/en/latest/python/views.html
"""

import logging
import secrets
from http import HTTPStatus

import acme.jws
import josepy as jose
import pytz
from acme import messages

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509.oid import ExtensionOID

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import ValidationError
from django.db import transaction
from django.http import Http404
from django.http import HttpResponse
from django.http import JsonResponse
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from .. import ca_settings
from ..extensions import SubjectAlternativeName
from ..models import AcmeAccount
from ..models import AcmeAuthorization
from ..models import AcmeCertificate
from ..models import AcmeChallenge
from ..models import AcmeOrder
from ..models import CertificateAuthority
from ..subject import Subject
from ..tasks import acme_issue_certificate
from ..tasks import acme_validate_challenge
from ..tasks import run_task
from ..utils import validate_email
from .errors import AcmeBadCSR
from .errors import AcmeException
from .errors import AcmeMalformed
from .errors import AcmeUnauthorized
from .messages import NewOrder
from .responses import AcmeResponseAccount
from .responses import AcmeResponseAccountCreated
from .responses import AcmeResponseAuthorization
from .responses import AcmeResponseBadNonce
from .responses import AcmeResponseChallenge
from .responses import AcmeResponseError
from .responses import AcmeResponseForbidden
from .responses import AcmeResponseMalformed
from .responses import AcmeResponseMalformedPayload
from .responses import AcmeResponseNotFound
from .responses import AcmeResponseOrder
from .responses import AcmeResponseOrderCreated
from .responses import AcmeResponseUnauthorized
from .responses import AcmeResponseUnsupportedMediaType
from .utils import parse_acme_csr

log = logging.getLogger(__name__)


class AcmeDirectory(View):
    """
    `Equivalent LE URL <https://acme-v02.api.letsencrypt.org/directory>`__

    .. seealso:: `RFC 8555, section 7.1.1 <https://tools.ietf.org/html/rfc8555#section-7.1.1>`_
    """

    def _url(self, request, name, ca):  # pylint: disable=no-self-use
        return request.build_absolute_uri(reverse('django_ca:%s' % name, kwargs={'serial': ca.serial}))

    def get(self, request, serial=None):
        # pylint: disable=missing-function-docstring; standard Django view function
        if not ca_settings.CA_ENABLE_ACME:
            raise Http404('Page not found.')

        if serial is None:
            try:
                # NOTE: default() already calls usable()
                ca = CertificateAuthority.objects.acme().default()
            except ImproperlyConfigured:
                return AcmeResponseNotFound(message='No (usable) default CA configured.')
        else:
            try:
                # NOTE: Serial is already sanitized by URL converter
                ca = CertificateAuthority.objects.acme().usable().get(serial=serial)
            except CertificateAuthority.DoesNotExist:
                return AcmeResponseNotFound(message='%s: CA not found.' % serial)

        # Get some random data into the directory view, as explained in the Let's Encrypt directory:
        #   https://community.letsencrypt.org/t/adding-random-entries-to-the-directory/33417
        rnd = jose.encode_b64jose(secrets.token_bytes(16))

        directory = {
            rnd: "https://community.letsencrypt.org/t/adding-random-entries-to-the-directory/33417",
            "keyChange": "http://localhost:8000/django_ca/acme/todo/key-change",
            "newAccount": self._url(request, 'acme-new-account', ca),
            "newNonce": self._url(request, 'acme-new-nonce', ca),
            "newOrder": self._url(request, 'acme-new-order', ca),
            "revokeCert": "http://localhost:8000/django_ca/acme/todo/revoke-cert"
        }

        # Construct a "meta" object if and add it if any fields are defined. Note that the meta object is
        # optional (RFC 8555, section 7.1.1: "The object MAY additionally contain a "meta" field.").
        meta = {}
        if ca.website:
            meta['website'] = ca.website
        if ca.terms_of_service:
            meta['termsOfService'] = ca.terms_of_service
        if ca.caa_identity:
            meta['caaIdentities'] = [ca.caa_identity]  # array of string
        if meta:
            directory['meta'] = meta

        return JsonResponse(directory)


class AcmeGetNonceViewMixin:
    """View mixin that provides methods to get and validate a Nonce.

    Note that thix mixin depends on the presence of a ``serial`` argument to the URL resolver.
    """

    nonce_length = 32
    """Length of generated Nonces."""

    def get_nonce(self):
        """Get a random Nonce and add it to the cache."""

        data = secrets.token_bytes(self.nonce_length)
        nonce = jose.encode_b64jose(data)
        cache_key = 'acme-nonce-%s-%s' % (self.kwargs['serial'], nonce)
        cache.set(cache_key, 0)
        return nonce

    def validate_nonce(self, nonce):
        """Validate that the given nonce was issued and was not used before."""
        cache_key = 'acme-nonce-%s-%s' % (self.kwargs['serial'], nonce)
        try:
            count = cache.incr(cache_key)
        except ValueError:
            return False

        if count > 1:  # nonce was already used
            # NOTE: "incr" returns the *new* value, so "1" is the expected value.
            return False

        return True


@method_decorator(csrf_exempt, name='dispatch')
class AcmeBaseView(AcmeGetNonceViewMixin, View):
    """Base class for all ACME views."""

    ignore_body = False  # True if we want to ignore the message body
    post_as_get = False  # True if this is a POST-as-GET request (see RFC 8555, 6.3).
    requires_key = False  # True if we require a full key (-> new accounts)
    message_cls = None  # Set to class parsing the request body if post_as_get=False.

    def acme_request(self, **kwargs):
        """Function to handle the given ACME request. Views are expected to implement this function.

        Parameters
        ----------

        **kwargs : dict
            The arguments as passed by the resolver. If ``post_as_get`` is True, an instance of the class
            specified in ``message_cls`` is passed as the ``message`` keyword argument.
        """
        raise NotImplementedError  # pragma: no cover

    def get_message_cls(self, request, **kwargs):
        """Get the message class used for parsing the request body."""
        # pylint: disable=unused-argument; kwargs is not usually used
        return self.message_cls

    def set_link_relations(self, response, **kwargs):
        """Set Link releations headers according to RFC8288.

        `RFC8555, section 7.1 <https://tools.ietf.org/html/rfc8555#section-7.1>`_ states:

            The "index" link relation is present on all resources other than the directory and indicates the
            URL of the directory.

        .. seealso:: https://tools.ietf.org/html/rfc8288
        """

        kwargs['index'] = reverse('django_ca:acme-directory', kwargs={'serial': self.kwargs['serial']})
        response['Link'] = ', '.join('<%s>;rel="%s"' % (self.request.build_absolute_uri(v), k)
                                     for k, v in kwargs.items())

    #def log_request(self):
    #    """Function for logging prepared requests."""
    #    if getattr(settings, 'LOG_ACME_REQUESTS', None):
    #        prepared_key = self.__class__.__name__
    #        prepared_path = os.path.join(settings.FIXTURES_DIR, 'prepared-acme-requests.json')
    #        prepared_data = {}
    #        if os.path.exists(prepared_path):
    #            with open(prepared_path) as stream:
    #                prepared_data = json.load(stream)
    #        if prepared_key not in prepared_data:
    #            prepared_data[prepared_key] = []
    #        prepared_data[prepared_key].append(self.prepared)
    #        with open(prepared_path, 'w') as stream:
    #            json.dump(prepared_data, stream, indent=4)

    def dispatch(self, request, *args, **kwargs):
        if not ca_settings.CA_ENABLE_ACME:
            raise Http404('Page not found.')

        #self.prepared = OrderedDict()  # pylint: disable=attribute-defined-outside-init
        try:
            response = super().dispatch(request, *args, **kwargs)
            self.set_link_relations(response)
        except Exception as ex:  # pylint: disable=broad-except
            log.exception(ex)
            response = AcmeResponseError(message='Internal server error')

        #self.log_request()

        # RFC 8555, section 6.7:
        # An ACME server provides nonces to clients using the HTTP Replay-Nonce header field, as specified in
        # Section 6.5.1.  The server MUST include a Replay-Nonce header field in every successful response to
        # a POST request and SHOULD provide it in error responses as well.
        response['replay-nonce'] = self.get_nonce()
        return response

    def post(self, request, serial, **kwargs):
        # pylint: disable=missing-function-docstring; standard Django view function
        # pylint: disable=attribute-defined-outside-init
        # pylint: disable=too-many-return-statements,too-many-branches; b/c of the many checks

        # TODO: RFC 8555, 6.2 has a nice list of things to check here that we don't yet fully cover
        if request.content_type != 'application/jose+json':
            # RFC 8555, 6.2:
            # "Because client requests in ACME carry JWS objects in the Flattened JSON Serialization, they
            # must have the Content-Type header field set to "application/jose+json".  If a request does not
            # meet this requirement, then the server MUST return a response with status code 415 (Unsupported
            # Media Type).
            return AcmeResponseUnsupportedMediaType()

        #self.prepared['body'] = json.loads(request.body.decode('utf-8'))

        try:
            self.jws = acme.jws.JWS.json_loads(request.body)
        except (jose.errors.DeserializationError, TypeError):
            return AcmeResponseMalformed(message='Could not parse JWS token.')

        combined = self.jws.signature.combined
        if combined.jwk and combined.kid:
            # 'The "jwk" and "kid" fields are mutually exclusive.  Servers MUST reject requests that contain
            # both.'
            return AcmeResponseMalformed(message='jwk and kid are mutually exclusive.')

        # Get certificate authority for this request
        try:
            self.ca = CertificateAuthority.objects.acme().usable().get(serial=serial)
        except CertificateAuthority.DoesNotExist:
            return AcmeResponseNotFound(message="The requested CA cannot be found.")

        if combined.jwk:
            if not self.requires_key:
                return AcmeResponseMalformed(message='Request requires a JWK key ID.')

            self.jwk = combined.jwk  # set JWK from request
        elif combined.kid:
            if self.requires_key:
                return AcmeResponseMalformed(message='Request requires a full JWK key.')

            # combined.kid is a full URL pointing to the account.
            try:
                account = AcmeAccount.objects.viewable().get(ca=self.ca, kid=combined.kid)
            except AcmeAccount.DoesNotExist:
                return AcmeResponseUnauthorized(message='Account not found.')
            if account.usable is False:
                return AcmeResponseUnauthorized(message='Account not usable.')
            #self.prepared['thumbprint'] = account.thumbprint
            #self.prepared['pem'] = account.pem
            #self.prepared['account_pk'] = account.pk

            self.jwk = jose.JWK.load(account.pem.encode('utf-8'))  # load JWK from database
            self.account = account
        else:
            # ... 'Either "jwk" (JSON Web Key) or "kid" (Key ID)'
            return AcmeResponseMalformed(message='JWS contained neither key nor key ID.')

        if len(self.jws.signatures) != 1:  # pragma: no cover
            # RFC 8555, 6.2: "The JWS MUST NOT have multiple signatures"
            return AcmeResponseMalformed(message='Multiple JWS signatures encountered.')

        # "The JWS Protected Header MUST include the following fields:...
        if not combined.alg or combined.alg == 'none':  # pragma: no cover
            # ... "alg"
            return AcmeResponseMalformed(message='No algorithm specified.')

        # Verify JWS signature
        try:
            if not self.jws.verify(self.jwk):
                return AcmeResponseMalformed(message='JWS signature invalid.')
        except Exception:  # pylint: disable=broad-except
            return AcmeResponseMalformed(message='JWS signature invalid.')

        #self.prepared['nonce'] = jose.encode_b64jose(combined.nonce)
        if not self.validate_nonce(jose.encode_b64jose(combined.nonce)):
            # ... "nonce"
            return AcmeResponseBadNonce()

        if combined.url != request.build_absolute_uri():
            # ... "url"
            # RFC 8555 is not really clear on the required response code, but merely says "If the two do not
            # match, then the server MUST reject the request as unauthorized."
            return AcmeResponseUnauthorized(message='URL does not match.')

        if self.post_as_get is True:
            if self.jws.payload != b'':
                return AcmeResponseMalformed(message='Non-empty payload in get-as-post request.')

            try:
                response = self.acme_request(**kwargs)
            except AcmeException as e:
                response = e.get_response()
        elif self.ignore_body is True:
            try:
                response = self.acme_request(**kwargs)
            except AcmeException as e:
                response = e.get_response()
        else:
            message_cls = self.get_message_cls(request, **kwargs)

            try:
                message = message_cls.json_loads(self.jws.payload)
            except jose.errors.DeserializationError:
                return AcmeResponseMalformedPayload()

            try:
                response = self.acme_request(message=message, **kwargs)
            except AcmeException as e:
                response = e.get_response()

        return response


class AcmeNewNonceView(AcmeGetNonceViewMixin, View):
    """View to retrieve a new nonce for replay protection.

    Equivalent LE URL: https://acme-v02.api.letsencrypt.org/acme/new-nonce

    .. seealso::

       * `RFC 8555, section 6.5 <https://tools.ietf.org/html/rfc8555#section-6.5>`_
       * `RFC 8555, section 7.2 <https://tools.ietf.org/html/rfc8555#section-7.2>`_
    """

    def dispatch(self, request, *args, **kwargs):
        if not ca_settings.CA_ENABLE_ACME:
            raise Http404('Page not found.')

        response = super().dispatch(request, *args, **kwargs)
        response['replay-nonce'] = self.get_nonce()

        # RFC 8555, section 7.2:
        #
        #   The server MUST include a Cache-Control header field with the "no-store" directive in responses
        response['cache-control'] = 'no-store'
        return response

    def head(self, request, serial):  # pylint: disable=no-self-use
        """Get a new Nonce with a HEAD request."""
        # pylint: disable=method-hidden; false positive - View.setup() sets head as property if not defined
        # pylint: disable=unused-argument; false positive - really used by AcmeGetNonceViewMixin
        return HttpResponse()

    def get(self, request, serial):
        """Get a new Nonce with a GET request.

        Note that certbot always does a HEAD request, but RFC 8555, section 7.2 mandates support for GET
        requests.
        """
        # pylint: disable=unused-argument; false positive - really used by AcmeGetNonceViewMixin
        return HttpResponse(status=HTTPStatus.NO_CONTENT)  # 204, unlike HEAD, which has 200


class AcmeNewAccountView(AcmeBaseView):
    """Implements endpoint for creating a new account, that is ``/acme/new-account``.

    This view is called when the ACME client tries to register a new account.

    .. seealso:: `RFC 8555, 7.3 <https://tools.ietf.org/html/rfc8555#section-7.3>`_
    """
    message_cls = messages.Registration
    requires_key = True

    def validate_contacts(self, message):  # pylint: disable=no-self-use
        """Validate the contact information for this message."""

        for contact in message.contact:
            if contact.startswith(messages.Registration.email_prefix):
                addr = contact[len(messages.Registration.email_prefix):]

                # RFC 8555, section 7.3
                #
                #   Clients MUST NOT provide a "mailto" URL in the "contact" field that contains "hfields"
                #   [RFC6068] or more than one "addr-spec" in the "to" component.

                # We rule out quoted local address fields, otherwise it's extremely hard to validate
                # email addresses.
                if addr.startswith('"'):
                    raise AcmeMalformed('invalidContact', 'Quoted local part in email is not allowed.')

                # Since the local part is not quoted, it cannot contain a ',' either, so a ',' means there
                # is more than one "addr-spec" in the "to" component. (see RFC 8555 quote above)
                if ',' in addr:
                    raise AcmeMalformed('invalidContact', 'More than one addr-spec is not allowed.')

                # Validate that there are no hfields in the address.
                # NOTE: ',' appears to be valid in the local part according to RFC 5322
                _local, domain = addr.split('@', 1)
                if '?' in domain:
                    raise AcmeMalformed('invalidContact', '%s: hfields are not allowed.' % domain)

                # Finally, verify that we're getting at least a valid domain.
                try:
                    validate_email(addr)
                except ValueError as ex:
                    raise AcmeMalformed('invalidContact', '%s: Not a valid email address.' % domain) from ex
            else:
                # RFC 8555, section 7.3
                #
                #   If the server rejects a contact URL for using an unsupported scheme, it MUST raise an
                #   error of type "unsupportedContact", ...
                raise AcmeMalformed('unsupportedContact', '%s: Unsupported address scheme.' % contact)

    def acme_request(self, message):  # pylint: disable=arguments-differ; more concrete here
        pem = self.jwk['key'].public_bytes(
            encoding=Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode('utf-8').strip()
        thumbprint = jose.encode_b64jose(self.jwk.thumbprint())

        # RFC 8555, section 7.3:
        #
        #   If this field is present with the value "true", then the server MUST NOT create a new account if
        #   one does not already exist.  This allows a client to look up an account URL based on an account
        #   key (see Section 7.3.1).
        if message.only_return_existing:
            try:
                account = AcmeAccount.objects.get(thumbprint=thumbprint, pem=pem)
                return AcmeResponseAccount(self.request, account)
            except AcmeAccount.DoesNotExist:
                # RFC 8555, section 7.3:
                #
                #   ... account does not exist, then the server MUST return an error response with status code
                #   400 (Bad Request) and type "urn:ietf:params:acme:error:accountDoesNotExist".
                return AcmeResponseMalformed(typ='accountDoesNotExist', message='Account does not exist.')

        # RFC 8555, section 7.3.1
        #
        #   If the server receives a newAccount request signed with a key for which it already has an account
        #   registered with the provided account key, then it MUST return a response with status code 200 (OK)
        #   and provide the URL of that account in the Location header field.
        try:
            # NOTE: Filter for thumbprint too b/c index for the field should speed up lookups.
            account = AcmeAccount.objects.get(thumbprint=thumbprint, pem=pem)
            return AcmeResponseAccount(self.request, account)
        except AcmeAccount.DoesNotExist:
            pass

        if self.ca.acme_requires_contact and not message.emails:
            # NOTE: RFC 8555 does not specify an error code in this case
            return AcmeResponseUnauthorized(message='Must provide at least one contact address.')

        # Make sure that contact addresses are valid
        self.validate_contacts(message)

        account = AcmeAccount(
            ca=self.ca,
            contact='\n'.join(message.contact),
            status=AcmeAccount.STATUS_VALID,
            terms_of_service_agreed=message.terms_of_service_agreed,
            thumbprint=thumbprint,
            pem=pem
        )
        account.set_kid(self.request)

        # Call full_clean() so that model validation can do its magic
        try:
            account.full_clean()
            account.save()
        except ValidationError:
            return AcmeResponseMalformed(message="Account cannot be stored.")

        #self.prepared['thumbprint'] = account.thumbprint
        #self.prepared['pem'] = account.pem
        #self.prepared['account_pk'] = account.pem

        # RFC 8555, section 7.3
        #
        #   The server returns this account object in a 201 (Created) response, with the account URL in a
        #   Location header field.
        #
        # AcmeResponseAccountCreated adds the Location field currently.
        return AcmeResponseAccountCreated(self.request, account)


class AcmeAccountView(AcmeBaseView):
    pass


class AcmeAccountOrdersView(AcmeBaseView):
    pass


class AcmeNewOrderView(AcmeBaseView):
    """Implements endpoint for applying for a new certificate, that is ``/acme/new-order``.

    This is the first request of an ACME client when requesting a new certificate.

    If the client receives a successful response, it will next fetch the authorizations listed in it, which
    are served by :py:class:`~django_ca.views.AcmeAuthorizationView`.

    ``certbot`` sends the :py:class:`~acme:acme.messages.NewOrder` message via
    :py:meth:`~acme:acme.client.ClientV2.new_order`.

    .. seealso:: `RFC 8555, 7.4 <https://tools.ietf.org/html/rfc8555#section-7.4>`_
    """
    message_cls = NewOrder

    @transaction.atomic
    def acme_request(self, message):  # pylint: disable=arguments-differ; more concrete here
        now = timezone.now()
        if timezone.is_naive(now):
            now = timezone.make_aware(now, timezone=pytz.utc)

        if message.not_before and message.not_before < now:
            return AcmeResponseMalformed(message='Certificate cannot be valid before now.')
        if message.not_after and message.not_after > now + ca_settings.ACME_MAX_CERT_VALIDITY:
            return AcmeResponseMalformed(message='Certificate cannot be valid that long.')
        if message.not_before and message.not_after and message.not_before > message.not_after:
            return AcmeResponseMalformed(message='notBefore must be before notAfter.')
        if not message.identifiers:
            return AcmeResponseMalformedPayload()

        not_before = message.not_before
        not_after = message.not_after

        if not settings.USE_TZ and message.not_before:
            not_before = timezone.make_naive(not_before)
        if not settings.USE_TZ and message.not_after:
            not_after = timezone.make_naive(not_after)

        # TODO: test if identifiers are acceptable
        order = AcmeOrder.objects.create(account=self.account, not_before=not_before, not_after=not_after)

        authorizations = [self.request.build_absolute_uri(authz.acme_url)
                          for authz in order.add_authorizations(message.identifiers)]

        expires = order.expires
        if timezone.is_naive(expires):  # acme.messages.Order requires a timezone-aware object
            expires = timezone.make_aware(expires, timezone=pytz.utc)

        response = AcmeResponseOrderCreated(
            authorizations=authorizations,
            expires=expires,
            finalize=self.request.build_absolute_uri(order.acme_finalize_url),
            identifiers=message.identifiers,
            not_after=message.not_after,
            not_before=message.not_before,
            status=order.status,
        )
        response['Location'] = self.request.build_absolute_uri(order.acme_url)
        return response


class AcmeOrderView(AcmeBaseView):
    """Implements endpoint for viewing an order, that is ``/acme/order/<slug>/``.

    A client calls this view after calling :py:class:`~django_ca.views.AcmeOrderFinalizeView`, presumably
    to retrieve final information about this certificate. This view itself is sparsely documented in RFC 8555.

    This view does seem to be a little redundant too, as the response is almost identical to the response the
    client received in the previous request. Nevertheless, certbot still calls this view and fails without it.

    .. seealso:: `RFC 8555, 7.4 <https://tools.ietf.org/html/rfc8555#section-7.4>`_
    """
    post_as_get = True

    def acme_request(self, slug):  # pylint: disable=arguments-differ; more concrete here
        try:
            order = AcmeOrder.objects.viewable().account(self.account).get(slug=slug)
        except AcmeOrder.DoesNotExist:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            return AcmeResponseUnauthorized()
        #self.prepared['order'] = order.slug

        expires = order.expires
        if timezone.is_naive(expires):  # acme.messages.Order requires a timezone-aware object
            expires = timezone.make_aware(expires, timezone=pytz.utc)

        authorizations = order.authorizations.all()
        if order.status in [AcmeOrder.STATUS_VALID, AcmeOrder.STATUS_INVALID]:
            # RFC 8555, section 7.1.3:
            #
            #   For final orders (in the "valid" or "invalid" state), the authorizations that were completed.
            authorizations = authorizations.filter(status=AcmeAuthorization.STATUS_VALID)

        cert_url = None
        try:
            cert = AcmeCertificate.objects.get(order=order)
            if cert.cert and order.status == AcmeOrder.STATUS_VALID:
                # WARNING: certbot (at least version 0.31.0) will try to fetch the certificate immediately if
                # we return the URL. That view will fail if the certificate is not yet issued, and certbot
                # fails with an error.
                # This behavior is independent of the status of the order, despite the fact this is the field
                # that should be used for this according to RFC 8555.
                cert_url = self.request.build_absolute_uri(cert.acme_url)
        except AcmeCertificate.DoesNotExist:
            pass

        response = AcmeResponseOrder(
            status=order.status,
            expires=expires,
            identifiers=tuple([{'type': a.type, 'value': a.value} for a in authorizations]),
            authorizations=tuple([self.request.build_absolute_uri(a.acme_url) for a in authorizations]),
            certificate=cert_url
        )
        response['Location'] = self.request.build_absolute_uri(order.acme_url)
        return response


class AcmeOrderFinalizeView(AcmeBaseView):
    """Implements endpoint for applying for certificate issuance, that is ``/acme/order/<slug>/finalize``.

    The client is supposed to call this URL to submit its CSR, once "it believes it has fulfilled the server's
    requirements".

    Note that in practice, the client can call this endpoint only once while the order is "ready". The
    endpoint returns an error if the order is not ready, and the call updates the state to "processing".

    .. seealso:: `RFC 8555, 7.4 <https://tools.ietf.org/html/rfc8555#section-7.4>`_
    """
    message_cls = messages.CertificateRequest

    def validate_csr(self, message, authorizations):  # pylint: disable=no-self-use
        """Parse and validate the CSR, returns the PEM as str."""

        # Note: Jose wraps the CSR in a josepy.util.ComparableX509, that has *no* public member methods.
        # The only public attribute or function is the wrapped object. We encode it back to get the regular
        # PEM.
        # Note that the CSR received here is not an actual PEM, see AcmeCertificate.parse_csr()
        csr = parse_acme_csr(message.encode('csr'))
        if csr.is_signature_valid is False:
            raise AcmeBadCSR(message='CSR signature is not valid.')

        # Do not accept MD5 or SHA1 signatures
        if isinstance(csr.signature_hash_algorithm, (hashes.MD5, hashes.SHA1)):
            raise AcmeBadCSR(message='%s: Insecure hash algorithm.' % csr.signature_hash_algorithm.name)

        # Get list of general names from the authorizations
        names_from_order = set(SubjectAlternativeName({
            'value': [auth.subject_alternative_name for auth in authorizations]
        }).extension_type)

        # Test if any subject Common Name is in the names for this order
        # NOTE: certbot does *not* set name in the subject
        csr_subject = Subject(csr.subject)
        if csr_subject.get('CN') and x509.DNSName(csr_subject.get('CN')) not in names_from_order:
            raise AcmeBadCSR(message='CommonName was not in order.')

        try:
            names_from_csr = set(
                csr.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value
            )
        except x509.ExtensionNotFound:
            raise AcmeBadCSR(message='No subject alternative names found in CSR.')

        if names_from_order != names_from_csr:
            raise AcmeBadCSR(message="Names in CSR do not match.")

        return csr.public_bytes(Encoding.PEM).decode('utf-8')

    def acme_request(self, message, slug):  # pylint: disable=arguments-differ; more concrete here
        try:
            order = AcmeOrder.objects.viewable().account(account=self.account).get(slug=slug)
        except AcmeOrder.DoesNotExist:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            return AcmeResponseUnauthorized()
        #self.prepared['order'] = order.slug

        # RFC 8555, section 7.4:
        #
        #   A request to finalize an order will result in error if the order is not in the "ready" state.  In
        #   such cases, the server MUST return a 403 (Forbidden) error with a problem document of type
        #   "orderNotReady".  The client should then send a POST-as-GET request to the order resource to
        #   obtain its current state.  The status of the order will indicate what action the client should
        #   take (see below).
        if order.status != AcmeOrder.STATUS_READY:
            # NOTE: The provision quoted above means that you will *always* get "orderNotReady", even if you
            # fetch this URL and the certificate has already been issued. We might consider returning the
            # order instead in this case.
            # The spec also says you should send a POST-as-GET after a certain time, if the order is in the
            # processing state, but it's not entirely clear if that request should go here or the normal order
            # resource.
            # Further investigation is on what LE and certbot do is needed here.
            return AcmeResponseForbidden(typ='orderNotReady', message='This order is not yet ready.')

        expires = order.expires
        if timezone.is_naive(expires):  # acme.messages.Order requires a timezone-aware object
            expires = timezone.make_aware(expires, timezone=pytz.utc)

        authorizations = order.authorizations.all()
        for auth in authorizations:
            if auth.status != AcmeAuthorization.STATUS_VALID:
                # This is a state that should never happen in practice, because the order is only marked as
                # ready once all authorizations are valid.
                return AcmeResponseForbidden(typ='orderNotReady', message='This order is not yet ready.')

        # Parse and validate the CSR
        csr = self.validate_csr(message, authorizations)

        # Create AcmeCertificate object (at this point without cert, as it hasn't been issued yet)
        cert = AcmeCertificate.objects.create(order=order, csr=csr)

        # Update the status of the order to "processing"
        order.status = AcmeOrder.STATUS_PROCESSING
        order.save()

        # start task only after commit, see:
        # https://docs.djangoproject.com/en/dev/topics/db/transactions/#django.db.transaction.on_commit
        transaction.on_commit(lambda: run_task(acme_issue_certificate, acme_certificate_pk=cert.pk))

        response = AcmeResponseOrder(
            status=order.status,
            expires=expires,
            identifiers=tuple([{'type': a.type, 'value': a.value} for a in authorizations]),
            authorizations=tuple([self.request.build_absolute_uri(a.acme_url) for a in authorizations]),
        )
        response['Location'] = self.request.build_absolute_uri(order.acme_url)
        return response


class AcmeCertificateView(AcmeBaseView):
    """Implements endpoint to download a certificate, that is ``/acme/cert/<slug>/``.

    This is the final view called in the certificate validation process and downloads the issued certificate.

    .. seealso:: `RFC8555, 8555, 7.4.2 <https://tools.ietf.org/html/rfc8555#section-7.4.2>`_
    """
    post_as_get = True

    def acme_request(self, slug):  # pylint: disable=arguments-differ; more concrete here
        try:
            cert = AcmeCertificate.objects.viewable().account(self.account).get(slug=slug)
        except AcmeCertificate.DoesNotExist:
            return AcmeResponseUnauthorized()

        #self.prepared['cert'] = slug
        #self.prepared['csr'] = cert.csr
        #self.prepared['order'] = cert.order.slug
        bundle = '\n'.join([c.pub.strip() for c in cert.cert.bundle])
        return HttpResponse(bundle, content_type='application/pem-certificate-chain')


class AcmeAuthorizationView(AcmeBaseView):
    """Implements endpoint for identifier authorization, that is ``/acme/authz/<slug>/``.

    This is the second request when a client requests a new certificate and represents an authorization
    request for one of the identifiers in the order. The URL for this view is returned by the
    :py:class:`~django_ca.views.AcmeNewOrderView`.

    This view returns URLs to the challenges served by ::py:class:`~django_ca.views.AcmeChallengeView`. The
    client will then post to one of these challenges to start the validation process.

    The client periodically polls this view after initiating a challenge to know when the server has
    successfully validated the challenge.

    Note that this resource accepts both a post-as-get request but also a response body: RFC 8555, section
    7.5.2 describes deactivating an account authorization.

    .. seealso:: `RFC 8555, 7.5 <https://tools.ietf.org/html/rfc8555#section-7.5>`_
    """

    post_as_get = True

    def acme_request(self, slug):  # pylint: disable=arguments-differ; more concrete here
        # TODO: implement deactivating an authorization (section 7.5.2)

        try:
            auth = AcmeAuthorization.objects.viewable().account(account=self.account).url().get(slug=slug)
        except AcmeAuthorization.DoesNotExist:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            return AcmeResponseUnauthorized()

        #self.prepared['order'] = auth.order.slug
        #self.prepared['auth'] = auth.slug
        challenges = auth.get_challenges()

        expires = auth.expires
        if not settings.USE_TZ:  # acme.Order requires a timezone-aware object
            expires = timezone.make_aware(expires, timezone=pytz.utc)

        # RFC8555, section 7.5.1:
        #
        #   "When finalizing an authorization, the server MAY remove challenges other than the one that was
        #   completed".
        #
        # The example response at the end of section 7.5.1 also only shows the valid challenge.
        if auth.status == AcmeAuthorization.STATUS_VALID:
            challenges = [c for c in challenges if c.status == AcmeChallenge.STATUS_VALID]

        resp = AcmeResponseAuthorization(
            identifier=auth.identifier,
            challenges=tuple([c.get_challenge(self.request) for c in challenges]),
            status=auth.status,
            expires=expires,
        )
        return resp


class AcmeChallengeView(AcmeBaseView):
    """Implements ``/acme/chall/<slug>``, indicating to the server that the challenge can now be validated.

    The client calls this view to tell the server to start validation of the resource of this challenges
    resource using the challenge method (http-01, dns-01, ...) for this challenge.

    After this view is called, the client will poll :py:class:`~django_ca.views.AcmeAuthorizationView` to know
    when the server has validated the challenge. After successful validation the client calls
    :py:class:`~django_ca.views.AcmeOrderFinalizeView` to request issuing of the certificate.

    .. seealso::

        * `RFC 8555, section 7.1.5 <https://tools.ietf.org/html/rfc8555#section-7.1.5>`_
        * `RFC 8555, section 7.5.1 <https://tools.ietf.org/html/rfc8555#section-7.5.1>`_
    """

    ignore_body = True

    def set_link_relations(self, response, **kwargs):
        """Set the "up" link header to the matching authorization.

        `RFC8555, section 7.1 <https://tools.ietf.org/html/rfc8555#section-7.1>`_ states:

            The "up" link relation is used with challenge resources to indicate the authorization resource to
            which a challenge belongs.
        """
        if response.status_code < HTTPStatus.BAD_REQUEST:
            # Only return an up relation if no error is thrown
            kwargs['up'] = self.auth.acme_url
        return super().set_link_relations(response, **kwargs)

    def acme_request(self, slug):  # pylint: disable=arguments-differ; more concrete here
        try:
            challenge = AcmeChallenge.objects.viewable().account(self.account).url().get(slug=slug)
        except AcmeChallenge.DoesNotExist:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            raise AcmeUnauthorized()  # pylint: disable=raise-missing-from

        # Set self.auth attribute, we need it in set_link_relations()
        self.auth = challenge.auth  # pylint: disable=attribute-defined-outside-init

        #self.prepared['order'] = challenge.auth.order.slug
        #self.prepared['auth'] = challenge.auth.slug
        #self.prepared['challenge'] = slug

        if challenge.usable is True:  # if not -> no state change
            # RFC8555, Section 7.1.6:
            #
            #   They transition to the "processing" state when the client responds to the challenge
            challenge.status = AcmeChallenge.STATUS_PROCESSING
            challenge.save()

            # Actually perform challenge validation asynchronously
            # start task only after commit, see:
            # https://docs.djangoproject.com/en/2.2/topics/db/transactions/#django.db.transaction.on_commit
            transaction.on_commit(lambda: run_task(acme_validate_challenge, challenge.pk))

        return AcmeResponseChallenge(
            chall=challenge.acme_challenge,
            _url=self.request.build_absolute_uri(challenge.acme_url),
            status=challenge.status,
            validated=challenge.validated
        )



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to view details for a certificate.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from datetime import datetime

from ..base import CertCommand


class Command(CertCommand):  # pylint: disable=missing-class-docstring
    binary_output = True
    allow_revoked = True
    help = 'View a certificate. The "list_certs" command lists all known certificates.'

    def add_arguments(self, parser):
        parser.add_argument(
            '-n', '--no-pem', default=False, action='store_true',
            help='Do not output public certificate in PEM format.')
        parser.add_argument(
            '-e', '--extensions', default=False, action='store_true',
            help='Show all extensions, not just subjectAltName.')
        self.add_format(parser)
        super().add_arguments(parser)

    def handle(self, cert, **options):  # pylint: disable=arguments-differ
        self.stdout.write('Common Name: %s' % cert.cn)

        # self.stdout.write notBefore/notAfter
        self.stdout.write('Valid from: %s' % cert.not_before.strftime('%Y-%m-%d %H:%M'))
        self.stdout.write('Valid until: %s' % cert.not_after.strftime('%Y-%m-%d %H:%M'))

        # self.stdout.write status
        now = datetime.utcnow()
        if cert.revoked:
            self.stdout.write('Status: Revoked')
        elif cert.not_after < now:
            self.stdout.write('Status: Expired')
        elif cert.not_before > now:
            self.stdout.write('Status: Not yet valid')
        else:
            self.stdout.write('Status: Valid')

        # self.stdout.write extensions
        if options['extensions']:
            self.print_extensions(cert)
        else:
            san = cert.subject_alternative_name
            if san:
                self.print_extension(san)

        self.stdout.write('Watchers:')
        for watcher in cert.watchers.all():
            self.stdout.write('* %s' % watcher)

        self.stdout.write('Digest:')
        for algo in ['md5', 'sha1', 'sha256', 'sha512']:
            self.stdout.write('    %s: %s' % (algo, cert.get_digest(algo)))

        self.stdout.write('HPKP pin: %s' % cert.hpkp_pin)

        if not options['no_pem']:
            self.stdout.write('')
            self.stdout.write(cert.dump_certificate(options['format']))



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Some constants from the ACME and associated standards.

.. WARNING::

   Do **not** import anything from acme or josepy here. This module is imported by various modules in the main
   django-ca library. Importing acme/josepy would make them a non-optional dependency.
"""

import enum
import string

# WARNING: Do not import acme or josepy, see above.

# base64url alphabet is defined in RFC 4648, section 5:
#   https://tools.ietf.org/html/rfc4648#section-5
# Jose JWS defines that the padding character ('=') is stripped:
#   https://tools.ietf.org/html/draft-ietf-jose-json-web-signature-37#section-2
BASE64_URL_ALPHABET = string.ascii_letters + string.digits + '-_'


@enum.unique
class Status(enum.Enum):
    """Enum of possible statuses for ACME objects.

    Duplicates :py:class:`~acme:acme.messages.Status` to avoid required ``acme`` during model import.
    """
    UNKNOWN = 'unknown'
    PENDING = 'pending'
    PROCESSING = 'processing'
    VALID = 'valid'
    INVALID = 'invalid'
    REVOKED = 'revoked'
    READY = 'ready'
    DEACTIVATED = 'deactivated'
    EXPIRED = 'expired'  # NOTE: not present in acme 1.9.0


@enum.unique
class IdentifierType(enum.Enum):
    """Enum of possible identifier types.

    Duplicates :py:class:`~acme:acme.messages.IdentifierType` to avoid required ``acme`` during model import.
    """
    DNS = 'dns'  # equivalent to acme.messages.IDENTIFIER_FQDN



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

from .responses import AcmeResponseBadCSR
from .responses import AcmeResponseError
from .responses import AcmeResponseMalformed
from .responses import AcmeResponseUnauthorized


class AcmeException(Exception):
    response = AcmeResponseError

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.kwargs = kwargs

    def get_response(self):
        return self.response(*self.args, **self.kwargs)


class AcmeMalformed(AcmeException):
    response = AcmeResponseMalformed


class AcmeUnauthorized(AcmeException):
    response = AcmeResponseUnauthorized


class AcmeBadCSR(AcmeException):
    """Exception raised when a CSR is not acceptable."""

    response = AcmeResponseBadCSR



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not,
# see <http://www.gnu.org/licenses/>.

"""Django model managers."""

import pathlib

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import PrivateFormat
from cryptography.x509.oid import AuthorityInformationAccessOID

from django.core.files.base import ContentFile
from django.db import models
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.encoding import force_str

from . import ca_settings
from .extensions import Extension
from .extensions import IssuerAlternativeName
from .extensions import NameConstraints
from .profiles import Profile
from .profiles import profiles
from .signals import post_create_ca
from .signals import post_issue_cert
from .signals import pre_create_ca
from .subject import Subject
from .utils import ca_storage
from .utils import generate_private_key
from .utils import get_cert_builder
from .utils import int_to_hex
from .utils import parse_hash_algorithm
from .utils import validate_hostname
from .utils import validate_key_parameters


class CertificateManagerMixin:  # pylint: disable=too-few-public-methods
    """Mixin for model managers."""

    def get_common_extensions(self, issuer_url=None, crl_url=None, ocsp_url=None):
        """Add extensions potentially common to both CAs and certs."""
        # pylint: disable=no-self-use

        extensions = []
        if crl_url:
            urls = [x509.UniformResourceIdentifier(force_str(c)) for c in crl_url]
            dps = [x509.DistributionPoint(full_name=[c], relative_name=None, crl_issuer=None, reasons=None)
                   for c in urls]
            extensions.append((False, x509.CRLDistributionPoints(dps)))
        auth_info_access = []
        if ocsp_url:
            uri = x509.UniformResourceIdentifier(force_str(ocsp_url))
            auth_info_access.append(x509.AccessDescription(
                access_method=AuthorityInformationAccessOID.OCSP, access_location=uri))
        if issuer_url:
            uri = x509.UniformResourceIdentifier(force_str(issuer_url))
            auth_info_access.append(x509.AccessDescription(
                access_method=AuthorityInformationAccessOID.CA_ISSUERS, access_location=uri))
        if auth_info_access:
            extensions.append((False, x509.AuthorityInformationAccess(auth_info_access)))
        return extensions

    def _extra_extensions(self, builder, extra_extensions):  # pylint: disable=no-self-use
        for ext in extra_extensions:
            if isinstance(ext, x509.extensions.Extension):
                builder = builder.add_extension(ext.value, critical=ext.critical)
            elif isinstance(ext, Extension):
                builder = builder.add_extension(*ext.for_builder())
            else:
                raise ValueError('Cannot add extension of type %s' % type(ext).__name__)
        return builder


class CertificateAuthorityManager(CertificateManagerMixin, models.Manager):
    """Model manager for the CertificateAuthority model."""

    def init(self, name, subject, expires=None, algorithm=None, parent=None, default_hostname=None,
             pathlen=None, issuer_url=None, issuer_alt_name='', crl_url=None, ocsp_url=None,
             ca_issuer_url=None, ca_crl_url=None, ca_ocsp_url=None, name_constraints=None,
             password=None, parent_password=None, ecc_curve=None, key_type='RSA', key_size=None,
             extra_extensions=None, path='ca',
             caa='', website='', terms_of_service='', acme_enabled=False, acme_requires_contact=True):
        """Create a new certificate authority.

        Parameters
        ----------

        name : str
            The name of the CA. This is a human-readable string and is used for administrative purposes only.
        subject : dict or str or :py:class:`~django_ca.subject.Subject`
            Subject string, e.g. ``"/CN=example.com"`` or ``Subject("/CN=example.com")``. The value is
            actually passed to :py:class:`~django_ca.subject.Subject` if it is not already an instance of that
            class.
        expires : datetime, optional
            Datetime for when this certificate authority will expire, defaults to
            :ref:`CA_DEFAULT_EXPIRES <settings-ca-default-expires>`.
        algorithm : str or :py:class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm`, optional
            Hash algorithm used when signing the certificate, passed to
            :py:func:`~django_ca.utils.parse_hash_algorithm`. The default is the value of the
            :ref:`CA_DIGEST_ALGORITHM <settings-ca-digest-algorithm>` setting.
        parent : :py:class:`~django_ca.models.CertificateAuthority`, optional
            Parent certificate authority for the new CA. Passing this value makes the CA an intermediate
            authority.
        default_hostname : str, optional
            Override the url configured with :ref:`CA_DEFAULT_HOSTNAME <settings-ca-default-hostname>` with a
            different hostname. Set to ``False`` to disable the hostname.
        pathlen : int, optional
            Value of the path length attribute for the :py:class:`~django_ca.extensions.BasicConstraints`
            extension.
        issuer_url : str
            URL for the DER/ASN1 formatted certificate that is signing certificates.
        issuer_alt_name : :py:class:`~django_ca.extensions.IssuerAlternativeName` or str, optional
            IssuerAlternativeName used when signing certificates. If the value is not an instance of
            :py:class:`~django_ca.extensions.IssuerAlternativeName`, it will be passed as argument to
            the constructor of the class.
        crl_url : list of str, optional
            CRL URLs used for certificates signed by this CA.
        ocsp_url : str, optional
            OCSP URL used for certificates signed by this CA. The default is no value, unless
            :ref:`CA_DEFAULT_HOSTNAME <settings-ca-default-hostname>` is set.
        ca_issuer_url : str, optional
            URL for the DER/ASN1 formatted certificate that is signing this CA. For intermediate CAs, this
            would usually be the ``issuer_url`` of the parent CA.
        ca_crl_url : list of str, optional
            CRL URLs used for this CA. This value is only meaningful for intermediate CAs.
        ca_ocsp_url : str, optional
            OCSP URL used for this CA. This value is only meaningful for intermediate CAs. The default is
            no value, unless :ref:`CA_DEFAULT_HOSTNAME <settings-ca-default-hostname>` is set.
        name_constraints : list of lists or :py:class:`~django_ca.extensions.NameConstraints`
            List of names that this CA can sign and/or cannot sign. The value is passed to
            :py:class:`~django_ca.extensions.NameConstraints` if the value is not already an instance of that
            class.
        password : bytes, optional
            Password to encrypt the private key with.
        parent_password : bytes, optional
            Password that the private key of the parent CA is encrypted with.
        ecc_curve : str or EllipticCurve, optional
            The elliptic curve to use for ECC type keys, passed verbatim to
            :py:func:`~django_ca.utils.parse_key_curve`.
        key_type: str, optional
            The type of private key to generate, must be one of ``"RSA"``, ``"DSA"`` or ``"ECC"``, with
            ``"RSA"`` being the default.
        key_size : int, optional
            Integer specifying the key size, must be a power of two (e.g. 2048, 4096, ...). Defaults to
            the :ref:`CA_DEFAULT_KEY_SIZE <settings-ca-default-key-size>`, unused if ``key_type="ECC"``.
        extra_extensions : list of :py:class:`cg:cryptography.x509.Extension` or \
                :py:class:`django_ca.extensions.Extension`, optional
            An optional list of additional extensions to add to the certificate.
        path : str or pathlib.PurePath, optional
            Where to store the CA private key (default ``ca``).
        caa : str, optional
            CAA identity. Note that this field is not yet currently used.
        website : str, optional
            URL to a human-readable website.
        terms_of_service : str, optional
            URL to the terms of service for the CA.
        acme_enabled : bool, optional
            Set to ``True`` to enable ACME support for this CA.
        acme_requires_contact : bool, optional
            Set to ``False`` if you do not want to force clients to register with an email address.

        Raises
        ------

        ValueError
            For various cases of wrong input data (e.g. ``key_size`` not being the power of two).
        PermissionError
            If the private key file cannot be written to disk.
        """
        # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements
        # NOTE: Already verified by KeySizeAction, so these checks are only for when the Python API is used
        #       directly.
        key_size, key_type, ecc_curve = validate_key_parameters(key_size, key_type, ecc_curve)
        algorithm = parse_hash_algorithm(algorithm)

        # Normalize extensions to django_ca.extensions.Extension subclasses
        if not isinstance(subject, Subject):
            subject = Subject(subject)
        if issuer_alt_name and not isinstance(issuer_alt_name, IssuerAlternativeName):
            issuer_alt_name = IssuerAlternativeName({'value': [issuer_alt_name]})
        if crl_url is None:
            crl_url = []

        serial = x509.random_serial_number()
        hex_serial = int_to_hex(serial)

        if default_hostname is None:
            default_hostname = ca_settings.CA_DEFAULT_HOSTNAME

        # If there is a default hostname, use it to compute some URLs from that
        if default_hostname:
            default_hostname = validate_hostname(default_hostname, allow_port=True)
            if parent:
                root_serial = parent.serial
            else:
                root_serial = hex_serial

            # Set OCSP urls
            if not ocsp_url:
                ocsp_path = reverse('django_ca:ocsp-cert-post', kwargs={'serial': hex_serial})
                ocsp_url = 'http://%s%s' % (default_hostname, ocsp_path)
            if parent and not ca_ocsp_url:  # OCSP for CA only makes sense in intermediate CAs
                ocsp_path = reverse('django_ca:ocsp-ca-post', kwargs={'serial': root_serial})
                ca_ocsp_url = 'http://%s%s' % (default_hostname, ocsp_path)

            # Set issuer path
            issuer_path = reverse('django_ca:issuer', kwargs={'serial': root_serial})
            if parent and not ca_issuer_url:
                ca_issuer_url = 'http://%s%s' % (default_hostname, issuer_path)
            if not issuer_url:
                issuer_url = 'http://%s%s' % (default_hostname, issuer_path)

            # Set CRL URLs
            if not crl_url:
                crl_path = reverse('django_ca:crl', kwargs={'serial': hex_serial})
                crl_url = ['http://%s%s' % (default_hostname, crl_path)]
            if parent and not ca_crl_url:  # CRL for CA only makes sense in intermediate CAs
                ca_crl_path = reverse('django_ca:ca-crl', kwargs={'serial': root_serial})
                ca_crl_url = ['http://%s%s' % (default_hostname, ca_crl_path)]

        pre_create_ca.send(
            sender=self.model, name=name, key_size=key_size, key_type=key_type, algorithm=algorithm,
            expires=expires, parent=parent, subject=subject, pathlen=pathlen, issuer_url=issuer_url,
            issuer_alt_name=issuer_alt_name, crl_url=crl_url, ocsp_url=ocsp_url, ca_issuer_url=ca_issuer_url,
            ca_crl_url=ca_crl_url, ca_ocsp_url=ca_ocsp_url, name_constraints=name_constraints,
            password=password, parent_password=parent_password, extra_extensions=extra_extensions, caa=caa,
            website=website, terms_of_service=terms_of_service, acme_enabled=acme_enabled,
            acme_requires_contact=acme_requires_contact)

        private_key = generate_private_key(key_size, key_type, ecc_curve)
        public_key = private_key.public_key()

        builder = get_cert_builder(expires, serial=serial)
        builder = builder.public_key(public_key)
        subject = subject.name

        builder = builder.subject_name(subject)
        builder = builder.add_extension(x509.BasicConstraints(ca=True, path_length=pathlen), critical=True)
        builder = builder.add_extension(x509.KeyUsage(
            key_cert_sign=True, crl_sign=True, digital_signature=False, content_commitment=False,
            key_encipherment=False, data_encipherment=False, key_agreement=False, encipher_only=False,
            decipher_only=False), critical=True)

        subject_key_id = x509.SubjectKeyIdentifier.from_public_key(public_key)
        builder = builder.add_extension(subject_key_id, critical=False)

        if parent is None:
            builder = builder.issuer_name(subject)
            private_sign_key = private_key
            aki = x509.AuthorityKeyIdentifier.from_issuer_public_key(public_key)
        else:
            builder = builder.issuer_name(parent.x509.subject)
            private_sign_key = parent.key(parent_password)
            aki = parent.get_authority_key_identifier()
        builder = builder.add_extension(aki, critical=False)

        for critical, ext in self.get_common_extensions(ca_issuer_url, ca_crl_url, ca_ocsp_url):
            builder = builder.add_extension(ext, critical=critical)

        if name_constraints:
            if not isinstance(name_constraints, NameConstraints):
                name_constraints = NameConstraints(name_constraints)

            builder = builder.add_extension(*name_constraints.for_builder())

        if extra_extensions:
            builder = self._extra_extensions(builder, extra_extensions)

        certificate = builder.sign(private_key=private_sign_key, algorithm=algorithm,
                                   backend=default_backend())

        # Normalize extensions for create()
        crl_url = '\n'.join(crl_url)

        ca = self.model(name=name, issuer_url=issuer_url, issuer_alt_name=','.join(issuer_alt_name),
                        ocsp_url=ocsp_url, crl_url=crl_url, parent=parent, caa_identity=caa, website=website,
                        terms_of_service=terms_of_service, acme_enabled=acme_enabled,
                        acme_requires_contact=acme_requires_contact)
        ca.x509 = certificate

        if password is None:
            encryption = serialization.NoEncryption()
        else:
            encryption = serialization.BestAvailableEncryption(password)

        pem = private_key.private_bytes(encoding=Encoding.PEM, format=PrivateFormat.PKCS8,
                                        encryption_algorithm=encryption)

        # write private key to file
        path = path / pathlib.PurePath('%s.key' % ca.serial.replace(':', ''))
        ca.private_key_path = ca_storage.save(str(path), ContentFile(pem))
        ca.save()

        post_create_ca.send(sender=self.model, ca=ca)
        return ca


class CertificateManager(CertificateManagerMixin, models.Manager):
    """Model manager for the Certificate model."""

    def parse_csr(self, csr, csr_format):  # pylint: disable=no-self-use; TODO: move to utils
        """Parse a CSR in the given format."""

        if isinstance(csr, x509.CertificateSigningRequest):
            return csr
        if csr_format == Encoding.PEM:
            return x509.load_pem_x509_csr(force_bytes(csr), default_backend())
        if csr_format == Encoding.DER:
            return x509.load_der_x509_csr(force_bytes(csr), default_backend())

        raise ValueError('Unknown CSR format passed: %s' % csr_format)

    def create_cert(self, ca, csr, csr_format=Encoding.PEM, profile=None, autogenerated=None, **kwargs):
        """Create and sign a new certificate based on the given profile.

        Parameters
        ----------

        ca : :py:class:`~django_ca.models.CertificateAuthority`
            The certificate authority to sign the certificate with.
        csr : str or :py:class:`~cg:cryptography.x509.CertificateSigningRequest`
            A valid CSR. If not already a :py:class:`~cg:cryptography.x509.CertificateSigningRequest`, the
            format is given by the ``csr_format`` parameter.
        csr_format : :py:class:`~cg:cryptography.hazmat.primitives.serialization.Encoding`, optional
            The format of the CSR. The default is ``PEM``.
        profile : str or :py:class:`~django_ca.profiles.Profile`, optional
            The name of a profile or a manually created :py:class:`~django_ca.profiles.Profile` instance. If
            not given, the profile configured by :ref:`CA_DEFAULT_PROFILE <settings-ca-default-profile>` is
            used.
        autogenerated : bool, optional
            Override the profiles ``autogenerated`` flag.
        **kwargs
            All other keyword arguments are passed to :py:func:`Profiles.create_cert()
            <django_ca.profiles.Profile.create_cert>`.
        """

        if not isinstance(profile, Profile):
            profile = profiles[profile]

        csr = self.parse_csr(csr, csr_format=csr_format)
        cert = profile.create_cert(ca, csr, **kwargs)

        obj = self.model(ca=ca, csr=csr.public_bytes(Encoding.PEM).decode('utf-8'), profile=profile.name)
        obj.x509 = cert
        if autogenerated is None:
            obj.autogenerated = profile.autogenerated
        else:
            obj.autogenerated = autogenerated
        obj.save()

        post_issue_cert.send(sender=self.model, cert=obj)

        return obj


class AcmeAccountManager(models.Manager):
    """Model manager for :py:class:`~django_ca.models.AcmeAccount`."""


class AcmeOrderManager(models.Manager):
    """Model manager for :py:class:`~django_ca.models.AcmeOrder`."""


class AcmeAuthorizationManager(models.Manager):
    """Model manager for :py:class:`~django_ca.models.AcmeAuthorization`."""


class AcmeChallengeManager(models.Manager):
    """Model manager for :py:class:`~django_ca.models.AcmeChallenge`."""


class AcmeCertificateManager(models.Manager):
    """Model manager for :py:class:`~django_ca.models.AcmeCertificate`."""



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to regenerate keys used for OCSP signing.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from datetime import timedelta

from django.core.management.base import CommandError

from ... import ca_settings
from ...models import CertificateAuthority
from ...tasks import generate_ocsp_key
from ...tasks import run_task
from ...utils import add_colons
from ..base import BaseCommand
from ..base import ExpiresAction


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "Regenerate OCSP keys."

    def add_arguments(self, parser):
        parser.add_argument(
            'serial', nargs='*',
            help="Generate OCSP keys only for the given CA. If omitted, generate keys for all CAs.")

        parser.add_argument(
            '--expires', default=timedelta(days=2), action=ExpiresAction,
            help='Sign the certificate for DAYS days (default: %(default)s)')
        parser.add_argument(
            '--quiet', action='store_true', default=False, help='Do not output warnings.')

        self.add_algorithm(parser)
        self.add_key_size(parser)
        self.add_key_type(parser)
        self.add_ecc_curve(parser)
        self.add_password(parser)

        self.add_profile(
            parser, 'Override the profile used for generating the certificate. By default, "ocsp" is used.')

    def handle(self, **options):  # pylint: disable=arguments-differ
        serials = options['serial']
        profile = options['profile'] or 'ocsp'

        # Check if the profile exists. Note that this shouldn't really happen, since valid parameters match
        # existing profiles. The only case is when the user undefines the "ocsp" profile, which is the
        # default.
        if profile not in ca_settings.CA_PROFILES:
            raise CommandError('%s: Undefined profile.' % profile)

        if not serials:
            serials = CertificateAuthority.objects.all().order_by('serial').values_list('serial', flat=True)

        for serial in serials:
            serial = serial.replace(':', '').strip().upper()
            hr_serial = add_colons(serial)
            try:
                ca = CertificateAuthority.objects.get(serial=serial)
            except CertificateAuthority.DoesNotExist:
                self.stderr.write(self.style.ERROR('%s: Unknown CA.' % hr_serial))
                continue

            if not ca.key_exists:
                if options['quiet'] is False:  # pragma: no branch
                    # NOTE: coverage falsely identifies the above condition to always be false.
                    self.stderr.write(self.style.WARNING('%s: CA has no private key.' % hr_serial))

                continue

            run_task(
                generate_ocsp_key,
                ca.serial,
                profile=profile,
                expires=options['expires'],
                algorithm=options['algorithm'],
                key_size=options['key_size'],
                key_type=options['key_type'],
                ecc_curve=options['ecc_curve'],
                password=options['password'],
            )



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""URL configuration for this project."""

from django.conf import settings
from django.urls import path
from django.urls import register_converter

from . import ca_settings
from . import converters
from . import views

app_name = 'django_ca'

register_converter(converters.AcmeSlugConverter, 'acme')
register_converter(converters.Base64Converter, 'base64')
register_converter(converters.HexConverter, 'hex')
register_converter(converters.SerialConverter, 'serial')

urlpatterns = [
    path('issuer/<hex:serial>.der', views.GenericCAIssuersView.as_view(), name='issuer'),
    path('ocsp/<hex:serial>/cert/', views.GenericOCSPView.as_view(expires=3600), name='ocsp-cert-post'),
    path('ocsp/<hex:serial>/cert/<base64:data>', views.GenericOCSPView.as_view(expires=3600),
         name='ocsp-cert-get'),
    path('ocsp/<hex:serial>/ca/', views.GenericOCSPView.as_view(ca_ocsp=True, expires=86400),
         name='ocsp-ca-post'),
    path('ocsp/<hex:serial>/ca/<base64:data>', views.GenericOCSPView.as_view(ca_ocsp=True, expires=86400),
         name='ocsp-ca-get'),
    path('crl/<hex:serial>/', views.CertificateRevocationListView.as_view(), name='crl'),
    path('crl/ca/<hex:serial>/', views.CertificateRevocationListView.as_view(scope='ca'), name='ca-crl'),
]


if ca_settings.CA_ENABLE_ACME:
    from .acme import views as acme_views

    # NOTE: Some functions depend on the fact that ALL ACME urls have a "serial" kwarg
    urlpatterns += [
        path('acme/directory/', acme_views.AcmeDirectory.as_view(), name='acme-directory'),
        path('acme/directory/<serial:serial>/', acme_views.AcmeDirectory.as_view(), name='acme-directory'),
        path('acme/<serial:serial>/new-nonce/', acme_views.AcmeNewNonceView.as_view(), name='acme-new-nonce'),
        path('acme/<serial:serial>/new-account/', acme_views.AcmeNewAccountView.as_view(),
             name='acme-new-account'),
        path('acme/<serial:serial>/new-order/', acme_views.AcmeNewOrderView.as_view(), name='acme-new-order'),
        path('acme/<serial:serial>/acct/<acme:slug>/', acme_views.AcmeAccountView.as_view(),
             name='acme-account'),
        path('acme/<serial:serial>/acct/<acme:slug>/orders/', acme_views.AcmeAccountOrdersView.as_view(),
             name='acme-account-orders'),
        path('acme/<serial:serial>/order/<acme:slug>/', acme_views.AcmeOrderView.as_view(),
             name='acme-order'),
        path('acme/<serial:serial>/order/<acme:slug>/finalize/', acme_views.AcmeOrderFinalizeView.as_view(),
             name='acme-order-finalize'),
        path('acme/<serial:serial>/authz/<acme:slug>/', acme_views.AcmeAuthorizationView.as_view(),
             name='acme-authz'),
        path('acme/<serial:serial>/chall/<acme:slug>/', acme_views.AcmeChallengeView.as_view(),
             name='acme-challenge'),
        path('acme/<serial:serial>/cert/<acme:slug>/', acme_views.AcmeCertificateView.as_view(),
             name='acme-cert'),
    ]


for name, kwargs in getattr(settings, 'CA_OCSP_URLS', {}).items():
    kwargs.setdefault('ca', name)
    urlpatterns += [
        path('ocsp/%s/' % name, views.OCSPView.as_view(**kwargs), name='ocsp-post-%s' % name),
        path('ocsp/%s/<base64:data>' % name, views.OCSPView.as_view(**kwargs), name='ocsp-get-%s' % name)
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to view details for a certificate authority.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from ... import ca_settings
from ...utils import add_colons
from ...utils import ca_storage
from ..base import BaseCommand


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = 'View details of a certificate authority.'

    def add_arguments(self, parser):
        self.add_ca(parser, arg='ca', allow_disabled=True, allow_unusable=True)

    def handle(self, ca, **options):  # pylint: disable=arguments-differ
        try:
            path = ca_storage.path(ca.private_key_path)
        except NotImplementedError:
            # Will raise NotImplementedError if storage backend does not support path(), in which case we use
            # the relative path from the database.
            # https://docs.djangoproject.com/en/dev/ref/files/storage/#django.core.files.storage.Storage.path
            path = ca.private_key_path

        self.stdout.write('%s (%s):' % (ca.name, 'enabled' if ca.enabled else 'disabled'))
        self.stdout.write('* Serial: %s' % add_colons(ca.serial))

        if ca_storage.exists(ca.private_key_path):
            self.stdout.write('* Path to private key:\n  %s' % path)
        else:
            self.stdout.write('* Private key not available locally.')

        if ca.parent:
            self.stdout.write('* Parent: %s (%s)' % (ca.parent.name, ca.parent.serial))
        else:
            self.stdout.write('* Is a root CA.')

        children = ca.children.all()
        if children:
            self.stdout.write('* Children:')
            for child in children:
                self.stdout.write('  * %s (%s)' % (child.name, child.serial))
        else:
            self.stdout.write('* Has no children.')

        pathlen = ca.pathlen
        if pathlen is None:
            pathlen = 'unlimited'

        self.stdout.write('* Distinguished Name: %s' % ca.distinguished_name)
        self.stdout.write('* Maximum levels of sub-CAs (pathlen): %s' % pathlen)
        self.stdout.write('* HPKP pin: %s' % ca.hpkp_pin)

        if ca.website:
            self.stdout.write('* Website: %s' % ca.website)
        if ca.terms_of_service:
            self.stdout.write('* Terms of service: %s' % ca.terms_of_service)
        if ca.caa_identity:
            self.stdout.write('* CAA identity: %s' % ca.caa_identity)

        if ca_settings.CA_ENABLE_ACME:
            self.stdout.write('')
            self.stdout.write('ACMEv2 support:')
            self.stdout.write('* Enabled: %s' % ca.acme_enabled)
            if ca.acme_enabled:
                self.stdout.write('* Requires contact: %s' % ca.acme_requires_contact)

        self.stdout.write('')
        self.stdout.write('X509 v3 certificate extensions for CA:')

        self.print_extensions(ca)

        self.stdout.write('')
        self.stdout.write('X509 v3 certificate extensions for signed certificates:')
        self.stdout.write('* Certificate Revokation List (CRL): %s' % (ca.crl_url or None))
        self.stdout.write('* Issuer URL: %s' % (ca.issuer_url or None))
        self.stdout.write('* OCSP URL: %s' % (ca.ocsp_url or None))
        self.stdout.write('* Issuer Alternative Name: %s' % (ca.issuer_alt_name or None))
        self.stdout.write('\n%s' % ca.pub)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Collection of constants used by django-ca."""

from enum import Enum


class ReasonFlags(Enum):
    """An enumeration for CRL reasons.

    This enumeration is a copy of ``cryptography.x509.ReasonFlags``. We create a copy because any change
    in the enumeration would trigger a database migration, so up/downgrading cryptography might cause problems
    with your Django project.
    """

    unspecified = "unspecified"
    key_compromise = "keyCompromise"
    ca_compromise = "cACompromise"
    affiliation_changed = "affiliationChanged"
    superseded = "superseded"
    cessation_of_operation = "cessationOfOperation"
    certificate_hold = "certificateHold"
    privilege_withdrawn = "privilegeWithdrawn"
    aa_compromise = "aACompromise"
    remove_from_crl = "removeFromCRL"
